// This unit contains a high performance counter class which can be used for
// highly accurate timing
//
// taken from Project JEDI Code Library (JCL)
// The Initial Developers of the Original Code are Theo Bebekis and
// Marcel van Brakel.
// Portions created by Marcel van Brakel are Copyright (C) Marcel van Brakel.
// All Rights Reserved.
// Portions created by Theo Bebekis are Copyright (C) Theo Bebekis.
// All Rights Reserved.
//
// Contributor(s):
//  Theo Bebekis
//  Marcel van Brakel
//  Robert Marquardt (marquardt)
//  Matthias Thoma (mthoma)
//  Petr Vones (pvones)
//
unit qCounterU;

interface

uses
  Windows;


type
  THQCounter = class
  private
    fCounting: Boolean;
    fElapsedTime: Double;
    fOverhead: Int64;
    fFrequency: Int64;
    fStart: Int64;
    fStop: Int64;

  protected
    function GetRunElapsedTime (): Double;

  public
    constructor Create (const compensate: Boolean = true);

    procedure Start ();
    function Stop (): Double;

    function Wait (time: Double): Double;

    property Counting: Boolean read fCounting;
    property ElapsedTime: Double read fElapsedTime;
    property Overhead: Int64 read fOverhead;
    property RunElapsedTime: Double read GetRunElapsedTime;
  end;


implementation


constructor THQCounter.Create (const compensate: Boolean);
const
  iterations = 10000;
var
  count: Integer;
  tmpOverhead: Int64;
begin
  inherited Create();
  if not QueryPerformanceFrequency(fFrequency) then
  begin
    WriteLn('***ERROR: there''s no high performance counter...');
    Halt(1);
  end;

  fCounting := false; fOverhead := 0;
  if compensate then
  begin
    // Determine overhead associated with calling of the Start and Stop methods.
    // This allows the Stop method to compensate for it and return a more
    // accurate result. Thanks to John O'Harrow (john att elmcrest dott demon dott co dott uk)
    tmpOverhead := 0;
    for count := 0 to iterations do
    begin
      Start(); Stop();
      tmpOverhead := tmpOverhead+(fStop-fStart);
    end;
    fOverhead := round(tmpOverhead/iterations);
  end;

  fElapsedTime := 0;
end;

procedure THQCounter.Start ();
begin
  fCounting := true; fElapsedTime := 0;
  if not QueryPerformanceCounter(fStart) then
  begin
    WriteLn('***ERROR: there''s no high performance counter...');
    Halt(1);
  end;
end;

function THQCounter.Stop (): Double;
begin
  if not QueryPerformanceCounter(fStop) then
  begin
    WriteLn('***ERROR: there''s no high performance counter...');
    Halt(1);
  end;
  fCounting := false;
  fElapsedTime := (fStop-fStart-fOverhead)/fFrequency;
  result := fElapsedTime;
end;

function THQCounter.GetRunElapsedTime (): Double;
var
  timeNow: Int64;
begin
  if not QueryPerformanceCounter(timeNow) then
  begin
    WriteLn('***ERROR: there''s no high performance counter...');
    Halt(1);
  end;
  result := (timeNow-fStart-fOverhead)/fFrequency;
end;

function THQCounter.Wait (time: Double): Double;
begin
  Start();
  repeat
    if not QueryPerformanceCounter(fStop) then
    begin
      WriteLn('***ERROR: there''s no high performance counter...');
      Halt(1);
    end;
    fElapsedTime := (fStop-fStart-fOverhead)/fFrequency;
  until fElapsedTime >= time;

  result := fElapsedTime;
end;


end.
