// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
unit dspLdrzU;

interface

uses
  SysUtils, z80U, dspTZXU, dspEmuVarsU, dspEmuMiscU;


const
  ZXLDR_UNKNOWN = 0;
  ZXLDR_HEWSON = 1; // hewson back-counting loader
  ZXLDR_UNILOADER = 2; // Paul Holmes' UniLoader

  ZXEDGE_UNKNOWN = 0;
  ZXEDGE_STDLIKE = 1; // standard-like


{============================================================}
// non-standard loaders
{============================================================}

// load byte with custom timings
// returns FALSE on error
// std timings: zeroC: 855; oneC: 1710
function ZXTapeLoadByte (out b: Byte; zeroC, oneC: Integer): Boolean;

// detect known non-standard loaders
// return ZXLDR_xxx
function ZXTapeDetectLoader (): Integer;

// emulate known non-standard loader
// should be called immediately after ZXTapeDetectLoader()!
// returns FALSE on error (tape pos abd registers unchanged)
function ZXTapeFastLoad (loaderType: Integer): Boolean;


{============================================================}
// LD-EDGE(2)
{============================================================}

// detect known LD-EDGE(2) routine
// return ZXEDGE_xxx
function ZXTapeDetectEdge (): Integer;

// emulate known non-standard loader
// should be called immediately after ZXTapeDetectEdge()!
procedure ZXTapeDoEdge (edgeType: Integer);


{============================================================}
// standard loader
{============================================================}

// return tick of the next pulse, skip this pulse
// result<0:  end of tape
function ZXTapeGetNextPulse (): Integer;

// load byte with the standard timings
// returns FALSE on error
function ZXTapeLoadStdByte (out b: Byte): Boolean;

// emulate std LD-BYTES routine (see std. LD-BYTES registers)
// returns FALSE on error
function ZXEmulateStdLoadBytes (): Boolean;

// emulate standard LD-BYTES routine
// should be called only when 48 ROM is paged in and PC=$0564
procedure ZXTapeStdFastLoader ();


implementation


const
  TAPE_TRESHOLD = 16;


{============================================================}
// standard loader
{============================================================}

function ZXTapeGetNextPulse (): Integer;
// skip special pulses
begin
  result := -2;
  // skip special pulses
  while emuTapeInfo.tape.Position < emuTapeInfo.tape.Count do
  begin
    result := emuTapeInfo.tape.ReadPulse();
    if result <= -1000 then
    begin
      result := (-result)-1000;
      if result < emuTapeInfo.tape.Messages.Count then
      begin
        WriteLn('tape msg: ', emuTapeInfo.tape.Messages[result]);
        EmuInfoMessage(emuTapeInfo.tape.Messages[result], 4000);
      end;
      continue;
    end;
    if result >= -3 then break;
  end;
end;

// !!!FIXME: "stop48k" block checking
function ZXTapeLoadStdByte (out b: Byte): Boolean;
// result: ok?
var
  f, c, n: Integer;
begin
  result := false; b := 0;

  for n := 0 to 7 do
  begin
    repeat
      f := ZXTapeGetNextPulse(); c := f;
    until f >= -3;
    if f < 0 then begin EmuStopTape(false); exit; end;
    if (abs(f-855) <= TAPE_TRESHOLD) or (abs(f-1710) <= TAPE_TRESHOLD) then
    begin
      // half-bit found, check the second half
      f := ZXTapeGetNextPulse();
      if f < 0 then begin EmuStopTape(false); exit; end;
      if abs(f-c) > TAPE_TRESHOLD then exit;
    end
    else exit; // error
    b := (b shl 1); if f > 1000 then b := b or 1;
  end;
  result := true;
end;

var
  saveAFx: Word;
  saveHL: Word;
  saveDE: Word;
  saveIX: Word;
  savePos: Integer;
  ldrmem: packed array [0..65535] of Byte;

procedure ActializeLdrMem ();
var
  w, a: Word;
begin
  w := saveDE; a := saveIX;
  repeat
    Z80PokeBInternal(emuZ80, a, ldrmem[a]);
    Inc(a);
    Dec(w);
  until w = 0;
end;

procedure SwapLdrMem ();
var
  w, a: Word;
  b: Byte;
begin
  w := saveDE; a := saveIX;
  repeat
    b := ldrmem[a];
    ldrmem[a] := Z80PeekBInternal(emuZ80, a);
    Z80PokeBInternal(emuZ80, a, b);
    Inc(a);
    Dec(w);
  until w = 0;
end;

procedure StdLoaderSaveRegs ();
begin
  saveAFx := emuZ80.regAFx;
  saveHL := emuZ80.regHL;
  saveDE := emuZ80.regDE;
  saveIX := emuZ80.regIX;
  savePos := emuTapeInfo.tape.Position;
end;

procedure StdLoaderRestoreRegs ();
begin
  emuZ80.regAFx := saveAFx;
  emuZ80.regHL := saveHL;
  emuZ80.regDE := saveDE;
  emuZ80.regIX := saveIX;
  emuTapeInfo.tape.Position := savePos;
end;

function ZXEmulateStdLoadBytes (): Boolean;
// result: ok?
label xxy;
var
  flag: Byte;
  f, c: Integer;
  b, checkB: Byte;
  w, a: Word;
begin
  result := false; StdLoaderSaveRegs();
  w := emuZ80.regDE; a := emuZ80.regIX;
  repeat
    ldrmem[a] := Z80PeekBInternal(emuZ80, a);
    Inc(a);
    Dec(w);
  until w = 0;

  flag := (emuZ80.regAFx shr 8) and $FF; // block type (flag)
  emuZ80.regHL := (emuZ80.regHL and $00FF) or $0100;
  // search the block type <flag>
  repeat
    checkB := 1; f := -2;
    // search for std. pilot
    while emuTapeInfo.tape.Position < emuTapeInfo.tape.Count do
    begin
      f := 10000; while (abs(f-2168) > TAPE_TRESHOLD) and (f > 0) do f := ZXTapeGetNextPulse();
      if f < 0 then begin EmuStopTape(false); exit; end;
      // flying %-)
      c := 0; while abs(f-2168) <= TAPE_TRESHOLD do begin Inc(c); f := ZXTapeGetNextPulse(); end;
      if f < 0 then begin EmuStopTape(false); exit; end;
      // pilot found (should be at least 1/2 sec), check sync1 & sync2
      if (c >= 807) and (abs(f-667) <= TAPE_TRESHOLD) and (abs(ZXTapeGetNextPulse()-735) <= TAPE_TRESHOLD) then break;
    end;
    if f < 0 then begin EmuStopTape(false); exit; end;
    if emuTapeInfo.tape.Position >= emuTapeInfo.tape.Count then exit; // tape ends

    if (emuZ80.regDE and $FF00) = $FF then begin checkB := 0; break; end;
    if ZXTapeLoadStdByte(checkB) then begin if checkB = flag then break; end; // found!
  until false;

  // now load the block
  emuZ80.regHL := (emuZ80.regHL and $00FF) or (checkB shl 8);
  if (emuZ80.regDE and $FF00) = $FF then goto xxy;
  repeat
    if not ZXTapeLoadStdByte(b) then exit; // tape error %-(
    // store byte
    //Z80PokeBInternal(emuZ80, emuZ80.regIX, b);
xxy:
    ldrmem[emuZ80.regIX] := b;
    // update parity
    emuZ80.regHL := (emuZ80.regHL and $00FF) or ((emuZ80.regHL and $FF00) xor (b shl 8));
    Inc(emuZ80.regIX);
    Dec(emuZ80.regDE);
  until emuZ80.regDE = 0;

  // now all bytes load ok, load the parity
  if not ZXTapeLoadStdByte(b) then exit; // tape error %-(
  // update parity byte
  emuZ80.regHL := (emuZ80.regHL and $00FF) or ((emuZ80.regHL and $FF00) xor (b shl 8));

  result := (emuZ80.regHL and $FF00) = 0;
end;

procedure ZXTapeStdFastLoader ();
// this procedure will be called when Speccy calls LD-BYTES
var
  bt: Byte;
  ok: Boolean;
begin
  if emuTapeInfo.tape = nil then exit;
  // if we are VERIFYing a tape, do not accelerate this (i'm so lazy! %-)
  if (emuZ80.regAFx and Z80FLAGC) = 0 then exit;
  //if (emuZ80.regDE = 0) or ((emuZ80.regDE and $FF00) = $FF00) then exit; // strange call %-)

  WriteLn('FastLoader(utm) invoked.');

  bt := (emuZ80.regAFx shr 8) and $FF; // block type (flag)
  Write('loading tape block type #', IntToHex(bt, 2), '...');

  ok := ZXEmulateStdLoadBytes();
  if not ok then
  begin
    StdLoaderRestoreRegs();
    if emuTapeInfo.tape.Position >= emuTapeInfo.tape.Count then
    begin
      WriteLn('failed!');
      WriteLn('no suitable tape blocks found in FastLoader(utm).');
      EmuStopTape(false);
      exit;
    end;
    WriteLn('tape error!');
    //if (emuZ80.regHL and $FF00) = 0 then emuZ80.regHL := (emuZ80.regHL and $00FF) or $0100;
  end
  else
  begin
    ActializeLdrMem();
    emuTapeInfo.nextEdgeT := emuZ80.tStates+32; //!!!???
    emuZ80.regPC := $5DF; // LD A,H; CP 1; RET
    WriteLn('ok');
  end;
end;


{============================================================}
// non-standard loaders
{============================================================}

// load byte with custom timings
// returns FALSE on error
// std timings: zeroC: 855; oneC: 1710
function ZXTapeLoadByte (out b: Byte; zeroC, oneC: Integer): Boolean;
var
  f, c, n: Integer;
begin
  result := false; b := 0;

  for n := 0 to 7 do
  begin
    repeat
      f := ZXTapeGetNextPulse(); c := f;
    until f >= -3;
    if f < 0 then begin EmuStopTape(false); exit; end;
    if (abs(f-855) <= TAPE_TRESHOLD) or (abs(f-1710) <= TAPE_TRESHOLD) then
    begin
      // half-bit found, check the second half
      f := ZXTapeGetNextPulse();
      if f < 0 then begin EmuStopTape(false); exit; end;
      if abs(f-c) > TAPE_TRESHOLD then exit;
    end
    else exit; // error
    b := (b shl 1); if f > 1000 then b := b or 1;
  end;
  result := true;
end;


function CheckByte (addr: Word; b: Byte): Boolean;
begin
  result := Z80PeekBInternal(emuZ80, addr) = b;
end;

function GetWord (addr: Word): Word;
begin
  result := Z80PeekBInternal(emuZ80, addr) or (Z80PeekBInternal(emuZ80, addr+1) shl 8);
end;

function CheckWord (addr, w: Word): Boolean;
var
  t: Word;
begin
  t := Z80PeekBInternal(emuZ80, addr) or (Z80PeekBInternal(emuZ80, addr+1) shl 8);
  result := t = w;
end;


var
  smc: Boolean; // loaded overwrites itself?

function IsHewson (): Boolean;
// detect hewson back-counting loader
var
  addr{, lde, lde2}: Word;
begin
  result := false;
  WriteLn('checking for HEWSON Loader...');

  if (emuZ80.regAFx and Z80FLAGC) = 0 then exit; // verify

  addr := emuZ80.regPC;
  if not CheckWord(addr-4, $FED3) then exit; // OUT (#FE),A

  if not CheckByte(addr-14, $14) then exit; // INC D
  if not CheckByte(addr-13, $08) then exit; // EX AF,AF'
  if not CheckByte(addr-12, $15) then exit; // DEC D
  if not CheckByte(addr-11, $F3) then exit; // DI
  if not CheckByte(addr-10, $D9) then exit; // EXX
  if not CheckByte(addr-7, $D9) then exit; // EXX

  if not CheckByte(addr+8, $CD) then exit; // CALL LD-EDGE
  if not CheckByte(addr+23, $CD) then exit; // CALL LD-EDGE2

  if not CheckByte(addr+128, $7C) then exit; // LD A,H
  if not CheckWord(addr+129, $01FE) then exit; // CP #01
  if not CheckByte(addr+131, $C9) then exit; // RET

  (*// address of LD-EDGE & LD-EDGE2
  lde := GetWord(addr+9); lde2 := GetWord(addr+24);
  if lde-lde2 <> 4 then exit;

  // check LD-EDGE
  if not CheckByte(lde2, $CD) then exit; // CALL
  if not CheckWord(lde2+1, lde) then exit;*)

  // check constants
  if not CheckWord(addr+28, $9C06) then exit; // LD B,$9C
  if not CheckWord(addr+35, $C63E) then exit; // LD A,$C6
  if not CheckWord(addr+43, $C906) then exit; // LD B,$C9
  if not CheckWord(addr+51, $D4FE) then exit; // CP $D4
  if not CheckWord(addr+70, $B006) then exit; // LD B,$B0
  if not CheckWord(addr+103, $B206) then exit; // LD B,$B2
  if not CheckWord(addr+111,$CB3E) then exit; // LD A,$CB
  if not CheckWord(addr+116,$B006) then exit; // LD B,$B0

  smc := (Integer(emuZ80.regIX)+Integer(emuZ80.regDE) >= addr);
  result := true;
  WriteLn('HEWSON loader found.');
end;

function IsUniLoader (): Boolean;
// detect Paul Holmes' UniLoader
var
  addr, lde, lde2: Word;
begin
  result := false;
  WriteLn('checking for UniLoader...');

  addr := emuZ80.regPC;
  if not CheckWord(addr-4, $FED3) then exit; // OUT (#FE),A

  if not CheckByte(addr-10,$14) then exit; // INC D
  if not CheckByte(addr-9, $08) then exit; // EX AF,AF'
  if not CheckByte(addr-8, $15) then exit; // DEC D
  if not CheckByte(addr-7, $F3) then exit; // DI

  if not CheckByte(addr+06, $CD) then exit; // CALL LD-EDGE
  if not CheckByte(addr+21, $CD) then exit; // CALL LD-EDGE2

  if not CheckByte(addr+116, $7C) then exit; // LD A,H
  if not CheckWord(addr+117, $01FE) then exit; // CP #01
  if not CheckByte(addr+119, $D8) then exit; // RET C

  // address of LD-EDGE & LD-EDGE2
  lde := GetWord(addr+7); lde2 := GetWord(addr+22);
  if lde-lde2 <> 4 then exit;

  // check LD-EDGE
  if not CheckByte(lde2, $CD) then exit; // CALL
  if not CheckWord(lde2+1, lde) then exit;
  if not CheckWord(lde, $163E) then exit; // LD A,$16

  // check constants
  if not CheckWord(addr+26, $9C06) then exit; // LD B,$9C
  if not CheckWord(addr+33, $C63E) then exit; // LD A,$C6
  if not CheckWord(addr+41, $C906) then exit; // LD B,$C9
  if not CheckWord(addr+49, $D4FE) then exit; // CP $D4
  if not CheckWord(addr+64, $B006) then exit; // LD B,$B0
  if not CheckWord(addr+91, $B206) then exit; // LD B,$B2
  if not CheckWord(addr+100,$CB3E) then exit; // LD A,$CB
  if not CheckWord(addr+105,$B006) then exit; // LD B,$B0

  smc := (Integer(emuZ80.regIX)+Integer(emuZ80.regDE) >= addr);
  result := true;
  WriteLn('UniLoader found.');
end;

// detect known non-standard loaders
// return ZXLDR_xxx
function ZXTapeDetectLoader (): Integer;
var
  addr: Word;
begin
  result := ZXLDR_UNKNOWN;
  //if (emuZ80.regDE = 0) or ((emuZ80.regDE and $FF00) = $FF00) then exit; // strange call %-)

  addr := emuZ80.regPC;
  if Z80PeekBInternal(emuZ80, addr) <> $1F then exit; // RRA?
  // previous of previous if OUT (#FE),A?
  if Z80PeekBInternal(emuZ80, addr-4) <> $D3 then exit;
  if Z80PeekBInternal(emuZ80, addr-3) <> $FE then exit;

  WriteLn('checking for known loaders');
  smc := false;
  if IsHewson() then result := ZXLDR_HEWSON
  else if IsUniLoader() then result := ZXLDR_UNILOADER;
end;

// emulate known non-standard loader
// should be called immediately after ZXTapeDetectLoader()!
function ZXTapeFastLoad (loaderType: Integer): Boolean;
var
  ofs, bc: Word;
  i, sp: Integer;
begin
  result := false;
  if emuTapeInfo.tape = nil then exit;
  if loaderType = ZXLDR_UNKNOWN then exit;
  case loaderType of
    ZXLDR_HEWSON, ZXLDR_UNILOADER:
      begin
        // save old tape position
        sp := emuTapeInfo.tape.Position;
        if not ZXEmulateStdLoadBytes() then
        begin
          StdLoaderRestoreRegs();
          // something's bad...
          exit;
        end;
        ofs := 0;
        // LD A,H; CP 1; RET
        if loaderType = ZXLDR_HEWSON then ofs := 128
        else if loaderType = ZXLDR_UNILOADER then ofs := 116;
        // self-modifying code?
        if smc then
        begin
          SwapLdrMem();
          i := ZXTapeDetectLoader();
          SwapLdrMem();
          if i <> loaderType then
          begin
            WriteLn('self-modifying loader!');
            StdLoaderRestoreRegs();
            exit;
          end;
        end;
        ActializeLdrMem();
        emuTapeInfo.nextEdgeT := emuZ80.tStates+32; //!!!???
        emuZ80.regPC := emuZ80.regPC+ofs;
        result := true;
      end;
  end;
end;


{============================================================}
// LD-EDGE(2)
{============================================================}

// detect known LD-EDGE(2) routine
// return ZXEDGE_xxx
function ZXTapeDetectEdge (): Integer;
begin
  result := ZXEDGE_UNKNOWN;
end;

// emulate known non-standard loader
// should be called immediately after ZXTapeDetectEdge()!
procedure ZXTapeDoEdge (edgeType: Integer);
begin
  if edgeType = ZXEDGE_UNKNOWN then exit;
end;


end.
