// Z80 CPU emulation engine v0.0.3
// based on vbSpec by Chris Cowley <ccowley@grok.co.uk>
// based on DelphiSpec v0.3 by Jari Korhonen <jarit.korhonen@luukku.com>
// havily changed by Ketmar // Vampire Avalon
//
// Complete Z80 emulation, including (as far as I know) the
// correct emulation of bits 3 and 5 of F, and undocumented ops.
// GNU GPL
//
// Ketmar notes:
//   this thing is not superfast (actually, it's superslow! %-), but at
//   least it works.
//   and i'm too lazy to write my own emulation engine from scratch.
//   anyway i've checked each opcode & write correct t-state changing (???)
//   so if U emulating contended memory, U can take 99.9% correct (lies!)
//   multicolor f/x. this slows down emulation a little, but who cares?
//
// changes:
//  [+] test set added (taken from FUSE)
//  [*] Z80 state moved in the separate record (so teoretically you can have
//      more than one working CPU %-)
//  [*] Peek/Poke/In/Out made "virtual", not built-in
//  [*] tStates counting in Z80 state record
//  [*] LDIR/LDDR changed (to allow self-modification & correct emulation at all)
//  [*] INI & company changed according to FUSE emulator
//  [*] OUTI & company changed according to FUSE emulator
//  [*] OUT (n),A corrected
//  [*] uniform "Integer" changed to correspondind data types
//  [*] BIT flag changing fixed (according to FUSE)
//  [*] flag3 & flag5 emulation in string commands (according to FUSE);
//      i'm unable to find another emulator (in sources) which does this...
//  [*] tStates incrementing corrected (to allow exact multicolor emulation)
//  [*] timing carefully checked (and it's still wrong! %-)
//  [+] custom opcode traps added
//  [*] stupid RLD/RRD bug fixed
//  [*] stupid flags bug fixed (missing byte to word conversions %-)
//  [*] stupid DAA flags bug fixed
//  [*] now DD/FD/xxx cannot be interrupted
//
// TODO:
//  [.] speed up LDI/LDD/CPI/CPD/...
//  [.] re-check timings
//
//  bug that WILL NOT be fixed: ID/ID/ID/.. sequence emulation.
//  this sequence should prevent interrupts while executing
//  (i.e. 17000 or so such prefixes forces Z80 to skip one intr)
//  but i see a little sense in such feature and never saw it used by someone.
//
{.$DEFINE USE_TRAPS}
{.$DEFINE ONLY_ED_TRAPS}
{.$DEFINE USE_ONEMOREREAD}
{.$DEFINE PACKED_REC}
{$ALIGN 4}
{$R-,S-}
unit z80U;

interface


// Flag bit positions
const
  Z80FLAGC = 1;
  Z80FLAGN = 2;
  Z80FLAGPV = 4;
  Z80FLAG3 = 8;
  Z80FLAGH = 16;
  Z80FLAG5 = 32;
  Z80FLAGZ = 64;
  Z80FLAGS = 128;


type
  PZ80CPU = ^TZ80CPU;
  TZ80CPU = {$IFDEF PACKED_REC}packed{$ENDIF} record
    // Main Z80 registers
    regA: Byte;
    regHL: Word;
    regB: Byte;
    regC: Byte;
    regDE: Word;

    // Z80 Flags
    flagS: Boolean;
    flagZ: Boolean;
    flag5: Boolean;
    flagH: Boolean;
    flag3: Boolean;
    flagPV: Boolean;
    flagN: Boolean;
    flagC: Boolean;

    // Alternate registers
    regAFx: Word;
    regHLx: Word;
    regBCx: Word;
    regDEx: Word;

    // index registers - ID used as temp for ix/iy
    regIX: Word;
    regIY: Word;
    regID: Word;

    // Stack pointer and program counter
    regSP: Word;
    regPC: Word;

    // Interrupt registers and flip-flops, and refresh registers
    regI: Byte;
    regR: Byte;  // high bit of R
    regRTemp: Byte; // low bits of R

    iff1: Boolean;
    iff2: Boolean;
    im: Byte;

    halted: Boolean;
      // controller must reset this flag in order to continue
      // executing (Z80 engine never resets it, only sets!)

    tStates: Integer; // t-state counter

    tStatesPerFrame: Integer; // used in LD A,I/LD A,R

    lastEItState: Integer; // should be zeroed at the frame start!

    prevPC: Word;
      // address of the 1st byte of the instruction that was executed
      // at the previous step

    orgPC: Word;
      // address of the 1st byte of the currently executing instruction

    userData: Pointer;
    breaked: Boolean;
      // set this to TRUE in PeekBExecProc to break Execute(). registers R, PC &
      // tStates value will not be changed by Execute() in this case.
      // actually, only Z80PeekBExec() should change this flag.

    {$IFDEF USE_ONEMOREREAD}
    oneMoreRead: Boolean;
      // the very internal thing used to emulate contended
      // memory. i just too lazy to explain it.
    {$ENDIF}
  end;


function Z80GetAF (z80cpu: PZ80CPU): Word;
function Z80GetF (z80cpu: PZ80CPU): Byte;

procedure Z80SetAF (z80cpu: PZ80CPU; v: Word);
procedure Z80SetF (z80cpu: PZ80CPU; b: Byte);
procedure Z80SetBC (z80cpu: PZ80CPU; nn: Word);
function Z80GetBC (z80cpu: PZ80CPU): Word;

function Z80Pop (z80cpu: PZ80CPU): Word; // uses PeekBInternalProc
procedure Z80Push (z80cpu: PZ80CPU; w: Word); // uses PokeBInternalProc

procedure Z80EXX (z80cpu: PZ80CPU);
procedure Z80ExAFAF (z80cpu: PZ80CPU);

procedure Z80Reset (z80cpu: PZ80CPU);

procedure Z80ClearTraps ();
procedure Z80ClearTrapsED ();

procedure Z80Execute (z80cpu: PZ80CPU; stoptStates: Integer);


var
  Z80PeekB: function (z80cpu: PZ80CPU; addr: Word): Byte = nil;
    // peek data byte (can change tStates)
  Z80PeekBExec: function (z80cpu: PZ80CPU): Byte = nil;
    // peek opcode byte (addr is z80cpu.regPC). can change tStates
    // this thing is separated from PeekB to allow
    // TR-DOS rom stamping or so...
    // if PeekBExecProc sets <breaked>, engine assumes that no
    // memory read & contention was happened, it
    // exits immediately w/o changing R, PC and tstate counter
  Z80PeekBInternal: function (z80cpu: PZ80CPU; addr: Word): Byte = nil;
    // peek data byte (CAN'T change tStates!)
  Z80PokeB: procedure (z80cpu: PZ80CPU; addr: Word; b: Byte) = nil;
    // poke data byte (can change tStates)
  Z80PokeBInternal: procedure (z80cpu: PZ80CPU; addr: Word; b: Byte) = nil;
    // poke data byte (CAN'T change tStates)
  Z80PortIn: function (z80cpu: PZ80CPU; port: Word): Byte = nil;
  Z80PortOut: procedure (z80cpu: PZ80CPU; port: Word; b: Byte) = nil;

  Z80Contention: procedure (z80cpu: PZ80CPU; addr: Word) = nil;
    // do memory contention w/o memory access

  Z80ED06Trap: procedure (z80cpu: PZ80CPU) = nil;
    // addr should be taken from z80cpu.regPC
    // it points to the byte right after #06

  Z80EDFBTrap: procedure (z80cpu: PZ80CPU) = nil;
    // .SLT trap
    //  HL: address to load;
    //  A: A --> level number
    //  return: CARRY complemented --> error
    // addr should be taken from z80cpu.regPC
    // it points to the byte right after #FB

  Z80SaveTrap: function (z80cpu: PZ80CPU): Boolean = nil;
    // return TRUE if trap processed and registers, etc. set ok

  {$IFDEF USE_TRAPS}
    {$IFNDEF ONLY_ED_TRAPS}
  Z80OpTraps: array [0..255] of function (z80cpu: PZ80CPU): Boolean;
    // when this is called, #ED already fetched, PC & R incremented by 1,
    // t-states incremented by 4 (+possible contention)
    // return FALSE to continue as usual
    // return TRUE to fetch & run new instruction
    // set <breaked> to break immediately (res=TRUE) or after
    // interpretation of the current opcode (res=FALSE).
    // no state unrolling will be made on immediate break
    {$ENDIF}
  Z80OpTrapsED: array [0..255] of function (z80cpu: PZ80CPU): Boolean;
    // when this is called, #ED already fetched, PC & R incremented by 1,
    // t-states incremented by 4 (+possible contention)
    // return FALSE to continue as usual
    // return TRUE to fetch & run new instruction
    // set <breaked> to break immediately (res=TRUE) or after
    // interpretation of the current opcode (res=FALSE).
    // no state unrolling will be made on immediate break
  {$ENDIF}


implementation


var
  z80Parity: array[0..256] of Boolean;
  //sz53Table: array [0.255] of Byte; (* The S, Z, 5 and 3 bits of the index *)


function Z80Pop (z80cpu: PZ80CPU): Word; // uses PeekBInternalProc
begin
  result := Z80PeekBInternal(z80cpu, z80cpu.regSP);
  Inc(z80cpu.regSP);
  Inc(result, Z80PeekBInternal(z80cpu, z80cpu.regSP)*256);
  Inc(z80cpu.regSP);
end;

procedure Z80Push (z80cpu: PZ80CPU; w: Word); // uses PokeBInternalProc
begin
  Dec(z80cpu.regSP);
  Z80PokeBInternal(z80cpu, z80cpu.regSP, (w and $FF00) shr 8);
  Dec(z80cpu.regSP);
  Z80PokeBInternal(z80cpu, z80cpu.regSP, w and $FF);
end;


function Z80Pop6T (z80cpu: PZ80CPU): Word;
begin
  Inc(z80cpu.tStates, 3); //SRL(3)
  result := Z80PeekB(z80cpu, z80cpu.regSP);
  Inc(z80cpu.regSP);
  Inc(z80cpu.tStates, 3); //SRH(3)
  Inc(result, Z80PeekB(z80cpu, z80cpu.regSP) shl 8);
  Inc(z80cpu.regSP);
end;

procedure Z80Push6T (z80cpu: PZ80CPU; w: Word);
begin
  Dec(z80cpu.regSP);
  Inc(z80cpu.tStates, 3); //SWH(3)
  Z80PokeB(z80cpu, z80cpu.regSP, (w and $FF00) shr 8);
  Dec(z80cpu.regSP);
  Inc(z80cpu.tStates, 3); //SWL(3)
  Z80PokeB(z80cpu, z80cpu.regSP, w and $FF);
end;


function LoadPCByteExec4T (z80cpu: PZ80CPU): Byte;
begin
  if z80cpu.breaked then begin result:= 0; exit; end;
  Inc(z80cpu.tStates, 4); //OCR(4)
  result := Z80PeekBExec(z80cpu);
  Inc(z80cpu.regPC);
end;

function LoadPCByte3T (z80cpu: PZ80CPU): Byte;
// Inc(tStates, 3);
begin
  Inc(z80cpu.tStates, 3); //ORD(3)
  result := Z80PeekB(z80cpu, z80cpu.regPC);
  Inc(z80cpu.regPC);
end;

function LoadPCWord6T (z80cpu: PZ80CPU): Word;
// Inc(tStates, 6);
begin
  Inc(z80cpu.tStates, 3); //ORL(3)
  result := Z80PeekB(z80cpu, z80cpu.regPC);
  Inc(z80cpu.regPC);
  Inc(z80cpu.tStates, 3); //ORH(3)
  Inc(result, Z80PeekB(z80cpu, z80cpu.regPC)*256);
  Inc(z80cpu.regPC);
end;

procedure PokeWord6T (z80cpu: PZ80CPU; addr, value: Word);
// Inc(tStates, 6);
begin
  Inc(z80cpu.tStates, 3); //MWL(3)
  Z80PokeB(z80cpu, addr, value and $FF);
  Inc(z80cpu.tStates, 3); //MWH(3)
  Z80PokeB(z80cpu, addr+1, value shr 8);
end;

function PeekWord6T (z80cpu: PZ80CPU; addr: Word): Word;
// Inc(tStates, 6);
begin
  Inc(z80cpu.tStates, 3); //ORL(3)
  result := Z80PeekB(z80cpu, addr);
  Inc(z80cpu.tStates, 3); //ORH(3)
  Inc(result, Z80PeekB(z80cpu, addr+1)*256);
end;


function Z80InBC (z80cpu: PZ80CPU): Word;
begin
  result := Z80PortIn(z80cpu, Z80GetBC(z80cpu));

  z80cpu.flagZ := (result = 0);
  z80cpu.flagS := (result and Z80FLAGS) <> 0;
  z80cpu.flag3 := (result and Z80FLAG3) <> 0;
  z80cpu.flag5 := (result and Z80FLAG5) <> 0;
  z80cpu.flagPV := z80Parity[result];
  z80cpu.flagN := false;
  z80cpu.flagH := false;
end;


procedure Z80AdcA (z80cpu: PZ80CPU; b: Byte);
var
  wans, ans, c: LongWord;
begin
  if z80cpu.flagC then c := 1 else c := 0;
  wans := Word(z80cpu.regA)+Word(b)+Word(c); ans := wans and $FF;
  z80cpu.flagS := (ans and Z80FLAGS) <> 0;
  z80cpu.flag3 := (ans and Z80FLAG3) <> 0;
  z80cpu.flag5 := (ans and Z80FLAG5) <> 0;
  z80cpu.flagZ := (ans = 0);
  z80cpu.flagC := (wans and $100) <> 0;
  z80cpu.flagPV := ((z80cpu.regA xor ((not b) and $FFFF)) and (z80cpu.regA xor ans) and $80) <> 0;
  z80cpu.flagH := (((z80cpu.regA and $0F)+(b and $0F)+c) and Z80FLAGH) <> 0;
  z80cpu.flagN := false;
  z80cpu.regA := ans;
end;

function Z80Adc16 (z80cpu: PZ80CPU; a, b: Word): Word;
var
  c, lans, ans: LongWord;
begin
  if z80cpu.flagC then c := 1 else c := 0;
  lans := LongWord(a)+LongWord(b)+LongWord(c); ans := lans and $FFFF;
  z80cpu.flagS := (ans and (Z80FLAGS*256)) <> 0;
  z80cpu.flag3 := (ans and (Z80FLAG3*256)) <> 0;
  z80cpu.flag5 := (ans and (Z80FLAG5*256)) <> 0;
  z80cpu.flagZ := (ans = 0);
  z80cpu.flagC := (lans and $10000) <> 0;
  z80cpu.flagPV := ((a xor ((not b) and $FFFF)) and (a xor ans) and $8000) <> 0;
  z80cpu.flagH := (((a and $0FFF)+(b and $0FFF)+c) and $1000) <> 0;
  z80cpu.flagN := false;
  result := ans;
end;

procedure Z80AddA (z80cpu: PZ80CPU; b: Byte);
var
  wans, ans: LongWord;
begin
  wans := Word(z80cpu.regA)+Word(b); ans := wans and $FF;
  z80cpu.flagS := (ans and Z80FLAGS) <> 0;
  z80cpu.flag3 := (ans and Z80FLAG3) <> 0;
  z80cpu.flag5 := (ans and Z80FLAG5) <> 0;
  z80cpu.flagZ := (ans = 0);
  z80cpu.flagC := (wans and $100) <> 0;
  z80cpu.flagPV := ((z80cpu.regA xor ((not (b)) and $FFFF)) and (z80cpu.regA xor ans) and $80) <> 0;
  z80cpu.flagH := (((z80cpu.regA and $0F)+(b and $0F)) and Z80FLAGH) <> 0;
  z80cpu.flagN := false;
  z80cpu.regA := ans;
end;

function Z80Add16 (z80cpu: PZ80CPU; a, b: Word): Word;
var
  lans, ans: LongWord;
begin
  lans := LongWord(a)+LongWord(b); ans := lans and $FFFF;
  z80cpu.flag3 := (ans and (Z80FLAG3*256)) <> 0;
  z80cpu.flag5 := (ans and (Z80FLAG5*256)) <> 0;
  z80cpu.flagC := (lans and $10000) <> 0;
  z80cpu.flagH := (((a and $FFF)+(b and $FFF)) and $1000) <> 0;
  z80cpu.flagN := false;
  result := ans;
end;

procedure Z80AndA (z80cpu: PZ80CPU; b: Byte);
begin
  z80cpu.regA := (z80cpu.regA and b);
  z80cpu.flagS := (z80cpu.regA and Z80FLAGS) <> 0;
  z80cpu.flag3 := (z80cpu.regA and Z80FLAG3) <> 0;
  z80cpu.flag5 := (z80cpu.regA and Z80FLAG5) <> 0;
  z80cpu.flagH := true;
  z80cpu.flagPV := z80Parity[z80cpu.regA];
  z80cpu.flagZ := (z80cpu.regA = 0);
  z80cpu.flagN := false;
  z80cpu.flagC := false;
end;

procedure Z80Bit (z80cpu: PZ80CPU; b, r: Word);
{var
  isBitSet: Boolean;}
begin
  (*
  isBitSet := (r and b) <> 0;
  z80cpu.flagN := false;
  z80cpu.flagH := true;
  ( *
  if b = Z80FLAG3 then z80cpu.flag3 := (r and Z80FLAG3) <> 0;
  if b = Z80FLAG5 then z80cpu.flag5 := (r and Z80FLAG5) <> 0;
  if b = Z80FLAGS then z80cpu.flagS := isBitSet else z80cpu.flagS := false;
  * )
  z80cpu.flag3 := (r and Z80FLAG3) <> 0;
  z80cpu.flag5 := (r and Z80FLAG5) <> 0;
  if b = Z80FLAGS then z80cpu.flagS := isBitSet else z80cpu.flagS := false;

  z80cpu.flagZ := not isBitSet;
  z80cpu.flagPV := z80cpu.flagZ;
  *)
  // slow, but correct %-)
  z80cpu.flagS := false;
  z80cpu.flagZ := false;
  z80cpu.flag5 := false;
  z80cpu.flagH := true;
  z80cpu.flag3 := false;
  z80cpu.flagPV := false;
  z80cpu.flagN := false;
  if (r and b) = 0 then begin z80cpu.flagPV := true; z80cpu.flagZ := true; end;
  if (b = $08) and ((r and $08) <> 0) then z80cpu.flag3 := true;
  if (b = $20) and ((r and $20) <> 0) then z80cpu.flag5 := true;
  if (b = $80) and ((r and $80) <> 0) then z80cpu.flagS := true;
end;

function Z80BitRes (z80cpu: PZ80CPU; bit, val: Word): Word;
begin
  result := val and (not (bit) and $FFFF);
end;

function Z80BitSet (z80cpu: PZ80CPU; bit, val: Word): Word;
begin
  result := val or bit;
end;

procedure Z80CCF (z80cpu: PZ80CPU);
begin
  z80cpu.flag3 := (z80cpu.regA and Z80FLAG3) <> 0;
  z80cpu.flag5 := (z80cpu.regA and Z80FLAG5) <> 0;
  z80cpu.flagH := z80cpu.flagC;
  z80cpu.flagN := false;
  z80cpu.flagC := not z80cpu.flagC;
end;

procedure Z80CPA (z80cpu: PZ80CPU; b: Byte);
var
  a, wans, ans: LongWord;
begin
  a := z80cpu.regA;
  wans := a-Word(b); ans := wans and $FF;
  z80cpu.flagS := (ans and Z80FLAGS) <> 0;
  z80cpu.flag3 := (b and Z80FLAG3) <> 0;
  z80cpu.flag5 := (b and Z80FLAG5) <> 0;
  z80cpu.flagN := true;
  z80cpu.flagZ := (ans = 0);
  z80cpu.flagC := (wans and $100) <> 0;
  z80cpu.flagH := (((a and $0F)-(b and $0F)) and Z80FLAGH) <> 0;
  z80cpu.flagPV := ((a xor b) and (a xor ans) and $80) <> 0;
end;

procedure Z80CPLA (z80cpu: PZ80CPU);
begin
  z80cpu.regA := (z80cpu.regA xor $FF) and $FF;
  z80cpu.flag3 := (z80cpu.regA and Z80FLAG3) <> 0;
  z80cpu.flag5 := (z80cpu.regA and Z80FLAG5) <> 0;
  z80cpu.flagH := true;
  z80cpu.flagN := true;
end;

procedure Z80SubA (z80cpu: PZ80CPU; b: Byte);
var
  a, wans, ans: LongWord;
begin
  a := z80cpu.regA;
  wans := a-Word(b); ans := wans and $FF;
  z80cpu.flagS := (ans and Z80FLAGS) <> 0;
  z80cpu.flag3 := (ans and Z80FLAG3) <> 0;
  z80cpu.flag5 := (ans and Z80FLAG5) <> 0;
  z80cpu.flagZ := (ans = 0);
  z80cpu.flagC := (wans and $100) <> 0;
  z80cpu.flagPV := ((a xor b) and (a xor ans) and $80) <> 0;
  z80cpu.flagH := (((a and $0F)-(b and $0F)) and Z80FLAGH) <> 0;
  z80cpu.flagN := true;
  z80cpu.regA := ans;
end;

procedure Z80DAAA (z80cpu: PZ80CPU);
var
  ans, incr: LongWord;
  carry: Boolean;
begin
  incr := 0; ans := z80cpu.regA; carry := z80cpu.flagC;

  if z80cpu.flagH or ((ans and $0F) > $09) then incr := incr or $06;
  if carry or (ans > $9F) then incr := incr or $60;
  if (ans > $8F) and ((ans and $0F) > $09) then incr := incr or $60;
  if ans > $99 then carry := true;
  if z80cpu.flagN then Z80SubA(z80cpu, incr) else Z80AddA(z80cpu, incr);

  ans := z80cpu.regA; z80cpu.flagC := carry; z80cpu.flagPV := z80Parity[ans];
end;

procedure Z80ExAFAF (z80cpu: PZ80CPU);
var
  t: Word;
begin
  t := Z80GetAF(z80cpu);
  Z80SetAF(z80cpu, z80cpu.regAFx);
  z80cpu.regAFx := t;
end;

function Z80RLC (z80cpu: PZ80CPU; ans: Word): Word;
var
  c: Boolean;
begin
  c := (ans and $80) <> 0;
  if c then ans := (ans*2) or $1 else ans := ans*2;
  ans := ans and $FF;
  z80cpu.flagS := (ans and Z80FLAGS) <> 0;
  z80cpu.flag3 := (ans and Z80FLAG3) <> 0;
  z80cpu.flag5 := (ans and Z80FLAG5) <> 0;
  z80cpu.flagZ := (ans = 0);
  z80cpu.flagPV := z80Parity[ans];
  z80cpu.flagH := false;
  z80cpu.flagN := false;
  z80cpu.flagC := c;
  result := ans;
end;

procedure Z80SetBC (z80cpu: PZ80CPU; nn: Word);
begin
  z80cpu.regB := (nn and $FF00) shr 8;
  z80cpu.regC := nn and $FF;
end;

procedure Z80SetD (z80cpu: PZ80CPU; l: Byte);
begin
  z80cpu.regDE := (Word(l)*256) or (z80cpu.regDE and $FF);
end;

function Z80GetD (z80cpu: PZ80CPU): Byte;
begin
  result := z80cpu.regDE shr 8;
end;

procedure Z80SetE (z80cpu: PZ80CPU; l: Byte);
begin
  z80cpu.regDE := (z80cpu.regDE and $FF00) or l;
end;

function Z80GetE (z80cpu: PZ80CPU): Byte;
begin
  result := z80cpu.regDE and $FF;
end;

procedure Z80SetH (z80cpu: PZ80CPU; l: Byte);
begin
  z80cpu.regHL := (Word(l)*256) or (z80cpu.regHL and $FF);
end;

procedure Z80SetL (z80cpu: PZ80CPU; l: Byte);
begin
  z80cpu.regHL := (z80cpu.regHL and $FF00) or l;
end;

procedure Z80SetF (z80cpu: PZ80CPU; b: Byte);
begin
  z80cpu.flagS := (b and Z80FLAGS) <> 0;
  z80cpu.flagZ := (b and Z80FLAGZ) <> 0;
  z80cpu.flag5 := (b and Z80FLAG5) <> 0;
  z80cpu.flagH := (b and Z80FLAGH) <> 0;
  z80cpu.flag3 := (b and Z80FLAG3) <> 0;
  z80cpu.flagPV := (b and Z80FLAGPV) <> 0;
  z80cpu.flagN := (b and Z80FLAGN) <> 0;
  z80cpu.flagC := (b and Z80FLAGC) <> 0;
end;

function Z80GetF (z80cpu: PZ80CPU): Byte;
begin
  result := 0;
  if z80cpu.flagS then Inc(result, Z80FLAGS);
  if z80cpu.flagZ then Inc(result, Z80FLAGZ);
  if z80cpu.flag5 then Inc(result, Z80FLAG5);
  if z80cpu.flagH then Inc(result, Z80FLAGH);
  if z80cpu.flag3 then Inc(result, Z80FLAG3);
  if z80cpu.flagPV then Inc(result, Z80FLAGPV);
  if z80cpu.flagN then Inc(result, Z80FLAGN);
  if z80cpu.flagC then Inc(result, Z80FLAGC);
end;

function Z80GetAF (z80cpu: PZ80CPU): Word;
begin
  result := (Word(z80cpu.regA)*256) or Z80GetF(z80cpu);
end;

procedure Z80SetAF (z80cpu: PZ80CPU; v: Word);
begin
  z80cpu.regA := (v and $FF00) shr 8;
  Z80SetF(z80cpu, v and $FF);
end;

function Z80GetBC (z80cpu: PZ80CPU): Word;
begin
  result := (z80cpu.regB*256) or z80cpu.regC;
end;

function Z80GetH (z80cpu: PZ80CPU): Byte;
begin
  result := z80cpu.regHL shr 8;
end;

function Z80GetIDH (z80cpu: PZ80CPU): Byte;
begin
  result := (z80cpu.regID shr 8) and $FF;
end;

function Z80GetIDL (z80cpu: PZ80CPU): Byte;
begin
  result := z80cpu.regID and $FF;
end;

function Z80GetL (z80cpu: PZ80CPU): Byte;
begin
  result := z80cpu.regHL and $FF;
end;

procedure Z80RRCA (z80cpu: PZ80CPU);
var
  c: Boolean;
begin
  c := (z80cpu.regA and $01) <> 0;
  if c then z80cpu.regA := (z80cpu.regA shr 1) or $80
  else z80cpu.regA := z80cpu.regA shr 1;
  z80cpu.flag3 := (z80cpu.regA and Z80FLAG3) <> 0;
  z80cpu.flag5 := (z80cpu.regA and Z80FLAG5) <> 0;
  z80cpu.flagN := false;
  z80cpu.flagH := false;
  z80cpu.flagC := c;
end;

function Z80RL (z80cpu: PZ80CPU; ans: Word): Word;
var
  c: Boolean;
begin
  c := (ans and $80) <> 0;
  if z80cpu.flagC then ans := (ans*2) or $1 else ans := ans*2;
  ans := ans and $FF;
  z80cpu.flagS := (ans and Z80FLAGS) <> 0;
  z80cpu.flag3 := (ans and Z80FLAG3) <> 0;
  z80cpu.flag5 := (ans and Z80FLAG5) <> 0;
  z80cpu.flagZ := (ans = 0);
  z80cpu.flagPV := z80Parity[ans];
  z80cpu.flagH := false;
  z80cpu.flagN := false;
  z80cpu.flagC := c;
  result := ans;
end;

procedure Z80RLA (z80cpu: PZ80CPU);
var
  ans: Word;
  c: Boolean;
begin
  ans := z80cpu.regA;
  c := (ans and $80) <> 0;
  if z80cpu.flagC then ans := (ans*2) or $1 else ans := ans*2;
  ans := ans and $FF;
  z80cpu.flag3 := (ans and Z80FLAG3) <> 0;
  z80cpu.flag5 := (ans and Z80FLAG5) <> 0;
  z80cpu.flagN := false;
  z80cpu.flagH := false;
  z80cpu.flagC := c;
  z80cpu.regA := ans;
end;

procedure Z80RLCA (z80cpu: PZ80CPU);
var
  c: Boolean;
begin
  c := (z80cpu.regA and $80) <> 0;
  if c then z80cpu.regA := (z80cpu.regA*2) or 1
  else z80cpu.regA := z80cpu.regA*2;
  z80cpu.regA := z80cpu.regA and $FF;
  z80cpu.flag3 := (z80cpu.regA and Z80FLAG3) <> 0;
  z80cpu.flag5 := (z80cpu.regA and Z80FLAG5) <> 0;
  z80cpu.flagN := false;
  z80cpu.flagH := false;
  z80cpu.flagC := c;
end;

function Z80RR (z80cpu: PZ80CPU; ans: Word): Word;
var
  c: Boolean;
begin
  c := (ans and $1) <> 0;
  if z80cpu.flagC then ans := (ans shr 1) or $80 else ans := ans shr 1;
  z80cpu.flagS := (ans and Z80FLAGS) <> 0;
  z80cpu.flag3 := (ans and Z80FLAG3) <> 0;
  z80cpu.flag5 := (ans and Z80FLAG5) <> 0;
  z80cpu.flagZ := (ans = 0);
  z80cpu.flagPV := z80Parity[ans];
  z80cpu.flagH := false;
  z80cpu.flagN := false;
  z80cpu.flagC := c;
  result := ans;
end;

procedure Z80RRA (z80cpu: PZ80CPU);
var
  ans: Word;
  c: Boolean;
begin
  ans := z80cpu.regA;
  c := (ans and $1) <> 0;
  if z80cpu.flagC then ans := (ans shr 1) or $80 else ans := ans shr 1;
  z80cpu.flag3 := (ans and Z80FLAG3) <> 0;
  z80cpu.flag5 := (ans and Z80FLAG5) <> 0;
  z80cpu.flagN := false;
  z80cpu.flagH := false;
  z80cpu.flagC := c;
  z80cpu.regA := ans;
end;

function Z80RRC (z80cpu: PZ80CPU; ans: Word): Word;
var
  c: Boolean;
begin
  c := (ans and $1) <> 0;
  if c then ans := (ans shr 1) or $80 else ans := ans shr 1;
  z80cpu.flagS := (ans and Z80FLAGS) <> 0;
  z80cpu.flag3 := (ans and Z80FLAG3) <> 0;
  z80cpu.flag5 := (ans and Z80FLAG5) <> 0;
  z80cpu.flagZ := (ans = 0);
  z80cpu.flagPV := z80Parity[ans];
  z80cpu.flagH := false;
  z80cpu.flagN := false;
  z80cpu.flagC := c;
  result := ans;
end;

procedure Z80RRDA (z80cpu: PZ80CPU);
var
  ans, t, q: Word;
begin
  ans := z80cpu.regA;
  Inc(z80cpu.tstates, 3); //MRD(3)
  t := Z80PeekB(z80cpu, z80cpu.regHL);
  Inc(z80cpu.tstates, 4); //IOP(4)
  q := (ans shl 4) or (t shr 4);
  Inc(z80cpu.tstates, 3); //MWR(3)
  Z80PokeB(z80cpu, z80cpu.regHL, (q and $FF));
  ans := ((ans and $F0) or (t and $0F)) and $FF;
  z80cpu.flagS := (ans and Z80FLAGS) <> 0;
  z80cpu.flag3 := (ans and Z80FLAG3) <> 0;
  z80cpu.flag5 := (ans and Z80FLAG5) <> 0;
  z80cpu.flagZ := (ans = 0);
  z80cpu.flagPV := z80Parity[ans];//!FUSE!
  z80cpu.flagH := false;
  z80cpu.flagN := false;
  z80cpu.regA := ans;
end;

procedure Z80RLDA (z80cpu: PZ80CPU);
var
  ans, t, q: LongWord;
begin
  ans := z80cpu.regA;
  Inc(z80cpu.tstates, 3); //MRD(3)
  t := Z80PeekB(z80cpu, z80cpu.regHL);
  Inc(z80cpu.tstates, 4); //IOP(4)
  q := (t shl 4) or (ans and $0F);
  Inc(z80cpu.tstates, 3); //MWR(3)
  Z80PokeB(z80cpu, z80cpu.regHL, (q and $FF));
  ans := ((ans and $F0) or (t shr 4)) and $FF;
  z80cpu.flagS := (ans and Z80FLAGS) <> 0;
  z80cpu.flag3 := (ans and Z80FLAG3) <> 0;
  z80cpu.flag5 := (ans and Z80FLAG5) <> 0;
  z80cpu.flagZ := (ans = 0);
  z80cpu.flagPV := z80Parity[ans];//!FUSE!
  z80cpu.flagH := false;
  z80cpu.flagN := false;
  z80cpu.regA := ans;
end;

procedure Z80SbcA (z80cpu: PZ80CPU; b: Byte);
var
  a, wans, ans, c: LongWord;
begin
  a := z80cpu.regA;
  if z80cpu.flagC then c := 1 else c:=0;
  wans := a-Word(b)-Word(c); ans := wans and $FF;
  z80cpu.flagS := (ans and Z80FLAGS) <> 0;
  z80cpu.flag3 := (ans and Z80FLAG3) <> 0;
  z80cpu.flag5 := (ans and Z80FLAG5) <> 0;
  z80cpu.flagZ := (ans = 0);
  z80cpu.flagC := (wans and $100) <> 0;
  z80cpu.flagPV := ((a xor b) and (a xor ans) and $80) <> 0;
  z80cpu.flagH := ((LongWord(a and $0F)-Word(b and $0F)-Word(c)) and Z80FLAGH) <> 0;
  z80cpu.flagN := true;
  z80cpu.regA := ans;
end;

function Z80Sbc16 (z80cpu: PZ80CPU; a, b: Word): Word;
var
  c, lans, ans: LongWord;
begin
  if z80cpu.flagC then c := 1 else c := 0;
  lans := LongWord(a)-Word(b)-Word(c); ans := lans and $FFFF;
  z80cpu.flagS := (ans and (Z80FLAGS*256)) <> 0;
  z80cpu.flag3 := (ans and (Z80FLAG3*256)) <> 0;
  z80cpu.flag5 := (ans and (Z80FLAG5*256)) <> 0;
  z80cpu.flagZ := (ans = 0);
  z80cpu.flagC := (lans and $10000) <> 0;
  z80cpu.flagPV := ((a xor b) and (a xor ans) and $8000) <> 0;
  z80cpu.flagH := ((LongWord(a and $0FFF)-Word(b and $0FFF)-Word(c)) and $1000) <> 0;
  z80cpu.flagN := true;
  result := ans;
end;

procedure Z80SCF (z80cpu: PZ80CPU);
begin
  z80cpu.flag3 := (z80cpu.regA and Z80FLAG3) <> 0;
  z80cpu.flag5 := (z80cpu.regA and Z80FLAG5) <> 0;
  z80cpu.flagN := false;
  z80cpu.flagH := false;
  z80cpu.flagC := true;
end;

function Z80SLA (z80cpu: PZ80CPU; ans: Word): Word;
var
  c: Boolean;
begin
  c := (ans and $80) <> 0;
  ans := (ans*2) and $FF;

  z80cpu.flagS := (ans and Z80FLAGS) <> 0;
  z80cpu.flag3 := (ans and Z80FLAG3) <> 0;
  z80cpu.flag5 := (ans and Z80FLAG5) <> 0;
  z80cpu.flagZ := (ans = 0);
  z80cpu.flagPV := z80Parity[ans];
  z80cpu.flagH := false;
  z80cpu.flagN := false;
  z80cpu.flagC := c;
  result := ans;
end;

function Z80SLS (z80cpu: PZ80CPU; ans: Word): Word;
var
  c: Boolean;
begin
  c := (ans and $80) <> 0;
  ans := ((ans*2) or $01) and $FF;
  z80cpu.flagS := (ans and Z80FLAGS) <> 0;
  z80cpu.flag3 := (ans and Z80FLAG3) <> 0;
  z80cpu.flag5 := (ans and Z80FLAG5) <> 0;
  z80cpu.flagZ := (ans = 0);
  z80cpu.flagPV := z80Parity[ans];
  z80cpu.flagH := false;
  z80cpu.flagN := false;
  z80cpu.flagC := c;
  result := ans;
end;

function Z80SRA (z80cpu: PZ80CPU; ans: Word): Word;
var
  c: Boolean;
begin
  c := (ans and $01) <> 0;
  ans := (ans shr 1) or (ans and $80);
  z80cpu.flagS := (ans and Z80FLAGS) <> 0;
  z80cpu.flag3 := (ans and Z80FLAG3) <> 0;
  z80cpu.flag5 := (ans and Z80FLAG5) <> 0;
  z80cpu.flagZ := (ans = 0);
  z80cpu.flagPV := z80Parity[ans];
  z80cpu.flagH := false;
  z80cpu.flagN := false;
  z80cpu.flagC := c;
  result := ans;
end;

function Z80SRL (z80cpu: PZ80CPU; ans: Word): Word;
var
  c: Boolean;
begin
  c := (ans and $1) <> 0;
  ans := ans shr 1;
  z80cpu.flagS := (ans and Z80FLAGS) <> 0;
  z80cpu.flag3 := (ans and Z80FLAG3) <> 0;
  z80cpu.flag5 := (ans and Z80FLAG5) <> 0;
  z80cpu.flagZ := (ans = 0);
  z80cpu.flagPV := z80Parity[ans];
  z80cpu.flagH := false;
  z80cpu.flagN := false;
  z80cpu.flagC := c;
  result := ans;
end;

procedure Z80XorA (z80cpu: PZ80CPU; b: Byte);
begin
  z80cpu.regA := (z80cpu.regA xor b);

  z80cpu.flagS := (z80cpu.regA and Z80FLAGS) <> 0;
  z80cpu.flag3 := (z80cpu.regA and Z80FLAG3) <> 0;
  z80cpu.flag5 := (z80cpu.regA and Z80FLAG5) <> 0;
  z80cpu.flagH := false;
  z80cpu.flagPV := z80Parity[z80cpu.regA];
  z80cpu.flagZ := (z80cpu.regA = 0);
  z80cpu.flagN := false;
  z80cpu.flagC := false;
end;

procedure Z80OrA (z80cpu: PZ80CPU; b: Byte);
begin
  z80cpu.regA := (z80cpu.regA or b);

  z80cpu.flagS := (z80cpu.regA and Z80FLAGS) <> 0;
  z80cpu.flag3 := (z80cpu.regA and Z80FLAG3) <> 0;
  z80cpu.flag5 := (z80cpu.regA and Z80FLAG5) <> 0;
  z80cpu.flagH := false;
  z80cpu.flagPV := z80Parity[z80cpu.regA];
  z80cpu.flagZ := (z80cpu.regA = 0);
  z80cpu.flagN := false;
  z80cpu.flagC := false;
end;

procedure Z80EXX (z80cpu: PZ80CPU);
var
  t: Word;
begin
  t := z80cpu.regHL;
  z80cpu.regHL := z80cpu.regHLx;
  z80cpu.regHLx := t;

  t := z80cpu.regDE;
  z80cpu.regDE := z80cpu.regDEx;
  z80cpu.regDEx := t;

  t := Z80GetBC(z80cpu);
  Z80SetBC(z80cpu, z80cpu.regBCx);
  z80cpu.regBCx := t;
end;

procedure Z80LdAI (z80cpu: PZ80CPU);
begin
  z80cpu.flagS := (z80cpu.regI and Z80FLAGS) <> 0;
  z80cpu.flag3 := (z80cpu.regI and Z80FLAG3) <> 0;
  z80cpu.flag5 := (z80cpu.regI and Z80FLAG5) <> 0;
  z80cpu.flagZ := (z80cpu.regI = 0);

  // taken from Unreal Spectrum
  z80cpu.flagPV := z80cpu.iff1 and (z80cpu.tStates+10 < z80cpu.tStatesPerFrame);
  //z80cpu.flagPV := z80cpu.iff1; // was iff2

  z80cpu.flagH := false;
  z80cpu.flagN := false;
  z80cpu.regA := z80cpu.regI;
end;

procedure Z80LdAR (z80cpu: PZ80CPU);
begin
  z80cpu.regRTemp := z80cpu.regRTemp and $7F;
  z80cpu.regA := (z80cpu.regR and $80) or z80cpu.regRTemp;
  z80cpu.flagS := (z80cpu.regA and Z80FLAGS) <> 0;
  z80cpu.flag3 := (z80cpu.regA and Z80FLAG3) <> 0;
  z80cpu.flag5 := (z80cpu.regA and Z80FLAG5) <> 0;
  z80cpu.flagZ := (z80cpu.regA = 0);

  // taken from Unreal Spectrum
  z80cpu.flagPV := z80cpu.iff2 and (z80cpu.tStates+10 < z80cpu.tStatesPerFrame);
  //z80cpu.flagPV := z80cpu.iff2;

  z80cpu.flagH := false;
  z80cpu.flagN := false;
end;

procedure Z80NegA (z80cpu: PZ80CPU);
var
  t: Word;
begin
  t := z80cpu.regA;
  z80cpu.regA := 0;
  Z80SubA(z80cpu, t);
end;

function Z80Inc8 (z80cpu: PZ80CPU; ans: Word): Word;
begin
  z80cpu.flagPV := (ans = $7F);
  z80cpu.flagH := (((ans and $0F)+1) and Z80FLAGH) <> 0;
  ans := (ans+1) and $FF;
  z80cpu.flagS := (ans and Z80FLAGS) <> 0;
  z80cpu.flag3 := (ans and Z80FLAG3) <> 0;
  z80cpu.flag5 := (ans and Z80FLAG5) <> 0;
  z80cpu.flagZ := (ans = 0);
  z80cpu.flagN := false;
  result := ans;
end;

function Z80Dec8 (z80cpu: PZ80CPU; ans: Word): Word;
begin
  z80cpu.flagPV := (ans = $80);
  z80cpu.flagH := (((ans and $0F)-1) and Z80FLAGH) <> 0;
  ans := (ans-1) and $FF;
  z80cpu.flagS := (ans and Z80FLAGS) <> 0;
  z80cpu.flag3 := (ans and Z80FLAG3) <> 0;
  z80cpu.flag5 := (ans and Z80FLAG5) <> 0;
  z80cpu.flagZ := (ans = 0);
  z80cpu.flagN := true;
  result := ans;
end;

function Z80IDD3T (z80cpu: PZ80CPU): Word;
var
  d: Integer;
begin
  d := LoadPCByte3T(z80cpu); //ORD(3)
  if (d and 128) = 128 then d := -(256-d);
  result := z80cpu.regID+d;
end;

procedure Z80SetIDH (z80cpu: PZ80CPU; byteval: Byte);
begin
  z80cpu.regID := ((byteval*256) and $FF00) or (z80cpu.regID and $FF);
end;

procedure Z80SetIDL (z80cpu: PZ80CPU; byteval: Byte);
begin
  z80cpu.regID := (z80cpu.regID and $FF00) or (byteval and $FF);
end;


procedure Z80ExecuteCB (z80cpu: PZ80CPU);
var
  op, n: Word;
begin
  op := LoadPCByte3T(z80cpu); Inc(z80cpu.tStates); //OCR(4)
  Inc(z80cpu.regRTemp);

  case op and $07 of
    0: n := z80cpu.regB;
    1: n := z80cpu.regC;
    2: n := (z80cpu.regDE shr 8) and $00FF;
    3: n := z80cpu.regDE and $00FF;
    4: n := (z80cpu.regHL shr 8) and $00FF;
    5: n := z80cpu.regHL and $00FF;
    6: // (HL)
      begin
        Inc(z80cpu.tStates, 4);//MRD(4)
        n := Z80PeekB(z80cpu, z80cpu.regHL);
      end;
    else n := z80cpu.regA;
  end;

  if op <= 7 then
  begin
    //RLC r
    n := Z80RLC(z80cpu, n) and $FF;
  end
  else if op <= 15 then
  begin
    //RRC r
    n := Z80RRC(z80cpu, n) and $FF;
  end
  else if op <= 23 then
  begin
    //RL r
    n := Z80RL(z80cpu, n) and $FF;
  end
  else if op <= 31 then
  begin
    //RR r
    n := Z80RR(z80cpu, n) and $FF;
  end
  else if op <= 39 then
  begin
    //SLA r
    n := Z80SLA(z80cpu, n) and $FF;
  end
  else if op <= 47 then
  begin
    //SRA r
    n := Z80SRA(z80cpu, n) and $FF;
  end
  else if op <= 55 then
  begin
    //SLL r
    n := Z80SLS(z80cpu, n) and $FF;
  end
  else if op <= 63 then
  begin
    //SRL r
    n := Z80SRL(z80cpu, n) and $FF;
  end
  else if op <= 127 then
  begin
    //BIT n,r
    Z80Bit(z80cpu, 1 shl ((op shr 3) and $07), n);
    op := 666;
  end
  else if op <= 191 then
  begin
    //RES n,r
    n := Z80BitRes(z80cpu, 1 shl ((op shr 3) and $07), n);
  end
  else
  begin
    //SET n,r
    n := Z80BitSet(z80cpu, 1 shl ((op shr 3) and $07), n);
  end;

  if op <> 666 then
  begin
    case op and $07 of
      0: z80cpu.regB := n;
      1: z80cpu.regC := n;
      2: z80cpu.regDE := (z80cpu.regDE and $00FF) or (n shl 8);
      3: z80cpu.regDE := (z80cpu.regDE and $FF00) or n;
      4: z80cpu.regHL := (z80cpu.regHL and $00FF) or (n shl 8);
      5: z80cpu.regHL := (z80cpu.regHL and $FF00) or n;
      6: // (HL)
        begin
          Inc(z80cpu.tStates, 3);//MWR(3)
          Z80PokeB(z80cpu, z80cpu.regHL, n);
        end;
      7: z80cpu.regA := n;
    end;
  end;
end;

procedure Z80ExecuteIDCB (z80cpu: PZ80CPU; op, z: Word);
var
  n: Word;
begin
  n := Z80PeekB(z80cpu, z); Inc(z80cpu.tStates, 3); //MRD(3)

  if op <= 7 then
  begin
    //RLC (ID+y) -> r
    n := Z80RLC(z80cpu, n) and $FF;
  end
  else if op <= 15 then
  begin
    //RRC (ID+y) -> r
    n := Z80RRC(z80cpu, n) and $FF;
  end
  else if op <= 23 then
  begin
    //RL (ID+y) -> r
    n := Z80RL(z80cpu, n) and $FF;
  end
  else if op <= 31 then
  begin
    //RR (ID+y) -> r
    n := Z80RR(z80cpu, n) and $FF;
  end
  else if op <= 39 then
  begin
    //SLA (ID+y) -> r
    n := Z80SLA(z80cpu, n) and $FF;
  end
  else if op <= 47 then
  begin
    //SRA (ID+y) -> r
    n := Z80SRA(z80cpu, n) and $FF;
  end
  else if op <= 55 then
  begin
    //SLL (ID+y) -> r
    n := Z80SLS(z80cpu, n) and $FF;
  end
  else if op <= 63 then
  begin
    //SRL (ID+y) -> r
    n := Z80SRL(z80cpu, n) and $FF;
  end
  else if op <= 127 then
  begin
    //BIT n,(ID+y)
    Z80Bit(z80cpu, 1 shl ((op shr 3) and $07), n);
    op := 666;
  end
  else if op <= 191 then
  begin
    //RES n,(ID+y) -> r
    n := Z80BitRes(z80cpu, 1 shl ((op shr 3) and $07), n);
  end
  else
  begin
    //SET n,(ID+y) -> r
    n := Z80BitSet(z80cpu, 1 shl ((op shr 3) and $07), n);
  end;

  if op <> 666 then
  begin
    case op and $07 of
      0: z80cpu.regB := n;
      1: z80cpu.regC := n;
      2: z80cpu.regDE := (z80cpu.regDE and $00FF) or (n shl 8);
      3: z80cpu.regDE := (z80cpu.regDE and $FF00) or n;
      4: z80cpu.regHL := (z80cpu.regHL and $00FF) or (n shl 8);
      5: z80cpu.regHL := (z80cpu.regHL and $FF00) or n;
      6: begin Inc(z80cpu.tStates, 3); Z80PokeB(z80cpu, z, n); end; //MWR(3) //???
      7: z80cpu.regA := n;
    end;
  end;

  Inc(z80cpu.tStates, 3); //MWR(3)
  Z80PokeB(z80cpu, z, n);
end;

procedure Z80ExecuteED (z80cpu: PZ80CPU);
var
  xxx: Byte;
  count, dest, from, b: Word;
  c: Boolean;
begin
  xxx := LoadPCByte3T(z80cpu); Inc(z80cpu.tStates); //OCR(4)
  Inc(z80cpu.regRTemp);
  case xxx of
    6: if Assigned(Z80ED06Trap) then Z80ED06Trap(z80cpu);
    $FB: if Assigned(Z80EDFBTrap) then Z80EDFBTrap(z80cpu);
    //0..5,7..63, 127..159, 164..167, 172..175, 180..183:;//NOP
    64: //IN B,(c)
      begin
        Inc(z80cpu.tStates, 4); //PRD(4)
        z80cpu.regB := Z80InBC(z80cpu);
      end;
    72: //IN C,(c)
      begin
        Inc(z80cpu.tStates, 4); //PRD(4)
        z80cpu.regC := Z80InBC(z80cpu);
      end;
    80: //IN D,(c)
      begin
        Inc(z80cpu.tStates, 4); //PRD(4)
        Z80SetD(z80cpu, Z80InBC(z80cpu));
      end;
    88: //IN E,(c)
      begin
        Inc(z80cpu.tStates, 4); //PRD(4)
        Z80SetE(z80cpu, Z80InBC(z80cpu));
      end;
    96: //IN H,(c)
      begin
        Inc(z80cpu.tStates, 4); //PRD(4)
        Z80SetH(z80cpu, Z80InBC(z80cpu));
      end;
    104: //IN L,(c)
      begin
        Inc(z80cpu.tStates, 4); //PRD(4)
        Z80SetL(z80cpu, Z80InBC(z80cpu));
      end;
    112: //IN (c)
      begin
        Inc(z80cpu.tStates, 4); //PRD(4)
        Z80InBC(z80cpu);
      end;
    120: //IN A,(c)
      begin
        Inc(z80cpu.tStates, 4); //PRD(4)
        z80cpu.regA := Z80InBC(z80cpu);
      end;
    65: //OUT (c),B
      begin
        Inc(z80cpu.tStates, 4); //PWR(4)
        Z80PortOut(z80cpu, Z80GetBC(z80cpu), z80cpu.regB);
      end;
    73: //OUT (c),C
      begin
        Inc(z80cpu.tStates, 4); //PWR(4)
        Z80PortOut(z80cpu, Z80GetBC(z80cpu), z80cpu.regC);
      end;
    81: //OUT (c),D
      begin
        Inc(z80cpu.tStates, 4); //PWR(4)
        Z80PortOut(z80cpu, Z80GetBC(z80cpu), (z80cpu.regDE shr 8));
      end;
    89: //OUT (c),E
      begin
        Inc(z80cpu.tStates, 4); //PWR(4)
        Z80PortOut(z80cpu, Z80GetBC(z80cpu), (z80cpu.regDE and $FF));
      end;
    97: //OUT (c),H
      begin
        Inc(z80cpu.tStates, 4); //PWR(4)
        Z80PortOut(z80cpu, Z80GetBC(z80cpu), (z80cpu.regHL shr 8));
      end;
    105: //OUT (c),L
      begin
        Inc(z80cpu.tStates, 4); //PWR(4)
        Z80PortOut(z80cpu, Z80GetBC(z80cpu), (z80cpu.regHL and $FF));
      end;
    113: //OUT (c),0
      begin
        Inc(z80cpu.tStates, 4); //PWR(4)
        Z80PortOut(z80cpu, Z80GetBC(z80cpu), 0);
      end;
    121: //OUT (c),A
      begin
        Inc(z80cpu.tStates, 4); //PWR(4)
        Z80PortOut(z80cpu, Z80GetBC(z80cpu), z80cpu.regA);
      end;
    66: //SBC HL,BC
      begin
        z80cpu.regHL := Z80Sbc16(z80cpu, z80cpu.regHL, Z80GetBC(z80cpu));
        Inc(z80cpu.tStates, 7); //IOP(4),IOP(3)
      end;
    74: //ADC HL,BC
      begin
        z80cpu.regHL := Z80Adc16(z80cpu, z80cpu.regHL, Z80GetBC(z80cpu));
        Inc(z80cpu.tStates, 7); //IOP(4),IOP(3)
      end;
    82: //SBC HL,DE
      begin
        z80cpu.regHL := Z80Sbc16(z80cpu, z80cpu.regHL, z80cpu.regDE);
        Inc(z80cpu.tStates, 7); //IOP(4),IOP(3)
      end;
    90: //ADC HL,DE
      begin
        z80cpu.regHL := Z80Adc16(z80cpu, z80cpu.regHL, z80cpu.regDE);
        Inc(z80cpu.tStates, 7); //IOP(4),IOP(3)
      end;
    98: //SBC HL,HL
      begin
        z80cpu.regHL := Z80Sbc16(z80cpu, z80cpu.regHL, z80cpu.regHL);
        Inc(z80cpu.tStates, 7); //IOP(4),IOP(3)
      end;
    106: //ADC HL,HL
      begin
        z80cpu.regHL := Z80Adc16(z80cpu, z80cpu.regHL, z80cpu.regHL);
        Inc(z80cpu.tStates, 7); //IOP(4),IOP(3)
      end;
    114: //SBC HL,SP
      begin
        z80cpu.regHL := Z80Sbc16(z80cpu, z80cpu.regHL, z80cpu.regSP);
        Inc(z80cpu.tStates, 7); //IOP(4),IOP(3)
      end;
    122: //ADC HL,SP
      begin
        z80cpu.regHL := Z80Adc16(z80cpu, z80cpu.regHL, z80cpu.regSP);
        Inc(z80cpu.tStates, 7); //IOP(4),IOP(3)
      end;
    67: //LD (nn),BC
      PokeWord6T(z80cpu, LoadPCWord6T(z80cpu), Z80GetBC(z80cpu));
    75: //LD BC,(nn)
      Z80SetBC(z80cpu, PeekWord6T(z80cpu, LoadPCWord6T(z80cpu)));
    83: //LD (nn),DE
      PokeWord6T(z80cpu, LoadPCWord6T(z80cpu), z80cpu.regDE);
    91: //LD DE,(nn)
      z80cpu.regDE := PeekWord6T(z80cpu, LoadPCWord6T(z80cpu));
    99: //LD (nn),HL
      PokeWord6T(z80cpu, LoadPCWord6T(z80cpu), z80cpu.regHL);
    107: //LD HL,(nn)
      z80cpu.regHL := PeekWord6T(z80cpu, LoadPCWord6T(z80cpu));
    115: //LD (nn),SP
      PokeWord6T(z80cpu, LoadPCWord6T(z80cpu), z80cpu.regSP);
    123: //LD SP,(nn)
      z80cpu.regSP := PeekWord6T(z80cpu, LoadPCWord6T(z80cpu));
    68, 76, 84, 92, 100, 108, 116, 124://NEG
        Z80NegA(z80cpu);
    69, 85, 101, 117: // RETN
      begin
        z80cpu.iff1 := z80cpu.iff2;
        z80cpu.regPC := Z80Pop6T(z80cpu);
      end;
    77, 93, 109, 125: //RETI
      begin
        // TOCHECK:
        //   according to the official Z80 docs, IFF2 does not get
        //   copied to IFF1 for RETI-but in a real Z80 it is
        z80cpu.iff1 := z80cpu.iff2;
        z80cpu.regPC := Z80Pop6T(z80cpu);
      end;
    70, 78, 102, 110: //IM 0
        z80cpu.im := 0;
    86, 118: //IM 1
        z80cpu.im := 1;
    94, 126: //IM 2
        z80cpu.im := 2;
    71: //LD I,A
      begin
        Inc(z80cpu.tStates); //OCR(5)
        z80cpu.regI := z80cpu.regA;
      end;
    79: //LD R,A
      begin
        Inc(z80cpu.tStates); //OCR(5)
        z80cpu.regR := z80cpu.regA;
        z80cpu.regRTemp := z80cpu.regR;
      end;
    87: //LD A,I
      begin
        Inc(z80cpu.tStates); //OCR(5)
        Z80LdAI(z80cpu);
      end;
    95: //LD A,R
      begin
        Inc(z80cpu.tStates); //OCR(5)
        Z80LdAR(z80cpu);
      end;
    103: //RRD
        Z80RRDA(z80cpu);
    111: //RLD
        Z80RLDA(z80cpu);
    160: //LDI
      begin
        Inc(z80cpu.tStates, 4);//MRD(3)
        b := Z80PeekB(z80cpu, z80cpu.regHL);
        Inc(z80cpu.tStates, 4);//MWR(3)
        Z80PokeB(z80cpu, z80cpu.regDE, b);
        Inc(z80cpu.regDE);
        Inc(z80cpu.regHL);
        Z80SetBC(z80cpu, Z80GetBC(z80cpu)-1);
        z80cpu.flagPV := (Z80GetBC(z80cpu) <> 0);
        z80cpu.flagH := false;
        z80cpu.flagN := false;
        //!FUSE!
        Inc(b, z80cpu.regA);
        z80cpu.flag3 := (b and Z80FLAG3) <> 0;
        z80cpu.flag5 := (b and $02) <> 0;
      end;
    161: //CPI
      begin
        Inc(z80cpu.tStates, 4);//MRD(3)
        b := Z80PeekB(z80cpu, z80cpu.regHL);
        Inc(z80cpu.tStates, 4);//MWR(3)
        Z80PokeB(z80cpu, z80cpu.regHL, b);
        c := z80cpu.flagC;
        Z80CPA(z80cpu, b);
        Inc(z80cpu.regHL);
        Z80SetBC(z80cpu, Z80GetBC(z80cpu)-1);
        z80cpu.flagPV := (Z80GetBC(z80cpu) <> 0);
        z80cpu.flagC := c;
        //!FUSE!
        b := z80cpu.regA-b; if z80cpu.flagH then Dec(b);
        z80cpu.flag3 := (b and Z80FLAG3) <> 0;
        z80cpu.flag5 := (b and $02) <> 0;
      end;
    162: //INI
      begin
        Inc(z80cpu.tStates);
        Inc(z80cpu.tStates, 4);//PRD(4)
        b := Z80PortIn(z80cpu, Z80GetBC(z80cpu));
        Inc(z80cpu.tStates, 3);//MWR(3)
        Z80PokeB(z80cpu, z80cpu.regHL, b);
        Dec(z80cpu.regB);
        Inc(z80cpu.regHL);
        //!FUSE!
        z80cpu.flagN := (b and $80) <> 0;
        b := z80cpu.regB;
        z80cpu.flagS := (b and $80) <> 0;
        z80cpu.flagZ := (b = 0);
        z80cpu.flag3 := (b and Z80FLAG3) <> 0;
        z80cpu.flag5 := (b and Z80FLAG5) <> 0;
        z80cpu.flagPV := false;
        z80cpu.flagH := false;
        z80cpu.flagC := false;
      end;
    163: //OUTI
      begin
        Inc(z80cpu.tStates);
        Dec(z80cpu.regB);
        Inc(z80cpu.tStates, 3);//MRD(3)
        b := Z80PeekB(z80cpu, z80cpu.regHL);
        Inc(z80cpu.tStates, 4);//PWR(4)
        Z80PortOut(z80cpu, Z80GetBC(z80cpu), b);
        Inc(z80cpu.regHL);
        //!FUSE!
        z80cpu.flagN := (b and $80) <> 0;
        b := z80cpu.regB;
        z80cpu.flagS := (b and $80) <> 0;
        z80cpu.flagZ := (b = 0);
        z80cpu.flag3 := (b and Z80FLAG3) <> 0;
        z80cpu.flag5 := (b and Z80FLAG5) <> 0;
        z80cpu.flagPV := false;
        z80cpu.flagH := false;
        z80cpu.flagC := false;
      end;
    168: //LDD
      begin
        Inc(z80cpu.tStates, 4);//MRD(3)
        b := Z80PeekB(z80cpu, z80cpu.regHL);
        Inc(z80cpu.tStates, 4);//MWR(3)
        Z80PokeB(z80cpu, z80cpu.regDE, b);
        Dec(z80cpu.regDE);
        Dec(z80cpu.regHL);
        Z80SetBC(z80cpu, Z80GetBC(z80cpu)-1);
        z80cpu.flagPV := (Z80GetBC(z80cpu) <> 0);
        z80cpu.flagH := false;
        z80cpu.flagN := false;
        //!FUSE!
        Inc(b, z80cpu.regA);
        z80cpu.flag3 := (b and Z80FLAG3) <> 0;
        z80cpu.flag5 := (b and $02) <> 0;
      end;
    169: //CPD
      begin
        Inc(z80cpu.tStates, 4);//MRD(3)
        b := Z80PeekB(z80cpu, z80cpu.regHL);
        Inc(z80cpu.tStates, 4);//MWR(3)
        Z80PokeB(z80cpu, z80cpu.regHL, b);
        c := z80cpu.flagC;
        Z80CPA(z80cpu, b);
        Dec(z80cpu.regHL);
        Z80SetBC(z80cpu, Z80GetBC(z80cpu)-1);
        z80cpu.flagPV := (Z80GetBC(z80cpu) <> 0);
        z80cpu.flagC := c;
        //!FUSE!
        b := z80cpu.regA-b; if z80cpu.flagH then Dec(b);
        z80cpu.flag3 := (b and Z80FLAG3) <> 0;
        z80cpu.flag5 := (b and $02) <> 0;
      end;
    170: //IND
      begin
        Inc(z80cpu.tStates);
        Inc(z80cpu.tStates, 4);//PRD(4)
        b := Z80PortIn(z80cpu, Z80GetBC(z80cpu));
        Inc(z80cpu.tStates, 3);//MWR(3)
        Z80PokeB(z80cpu, z80cpu.regHL, b);
        Dec(z80cpu.regB);
        Dec(z80cpu.regHL);
        //!FUSE!
        z80cpu.flagN := (b and $80) <> 0;
        b := z80cpu.regB;
        z80cpu.flagS := (b and $80) <> 0;
        z80cpu.flagZ := (b = 0);
        z80cpu.flag3 := (b and Z80FLAG3) <> 0;
        z80cpu.flag5 := (b and Z80FLAG5) <> 0;
        z80cpu.flagPV := false;
        z80cpu.flagH := false;
        z80cpu.flagC := false;
      end;
    171: //OUTD
      begin
        Inc(z80cpu.tStates);
        z80cpu.regB := Z80Dec8(z80cpu, z80cpu.regB);
        Inc(z80cpu.tStates, 3);//MRD(3)
        b := Z80PeekB(z80cpu, z80cpu.regHL);
        Inc(z80cpu.tStates, 4);//PWR(4)
        Z80PortOut(z80cpu, Z80GetBC(z80cpu), b);
        Dec(z80cpu.regHL);
        //!FUSE!
        z80cpu.flagN := (b and $80) <> 0;
        b := z80cpu.regB;
        z80cpu.flagS := (b and $80) <> 0;
        z80cpu.flagZ := (b = 0);
        z80cpu.flag3 := (b and Z80FLAG3) <> 0;
        z80cpu.flag5 := (b and Z80FLAG5) <> 0;
        z80cpu.flagPV := false;
        z80cpu.flagH := false;
        z80cpu.flagC := false;
      end;
    176: //LDIR
      begin
        count := Z80GetBC(z80cpu); dest := z80cpu.regDE; from := z80cpu.regHL;
        Inc(z80cpu.tStates, 3);//MRD(3)
        b := Z80PeekB(z80cpu, from);
        Inc(z80cpu.tStates, 5);//MWR(5)
        Z80PokeB(z80cpu, dest, b);
        Inc(from); Inc(dest); Dec(count);
        z80cpu.flagH := false;
        z80cpu.flagN := false;
        z80cpu.flagPV := (count <> 0);
        if count <> 0 then
        begin
          Dec(z80cpu.regPC, 2);
          Inc(z80cpu.tStates, 5); //IOP(5)
        end;
        z80cpu.regDE := dest; z80cpu.regHL := from; Z80SetBC(z80cpu, count);
        //!FUSE!
        Inc(b, z80cpu.regA);
        z80cpu.flag3 := (b and Z80FLAG3) <> 0;
        z80cpu.flag5 := (b and $02) <> 0;
      end;
    177: //CPIR
      begin
        c := z80cpu.flagC;
        Inc(z80cpu.tStates, 3);//MRD(3)
        b := Z80PeekB(z80cpu, z80cpu.regHL);
        Inc(z80cpu.tStates, 5);//MWR(5)
        Z80PokeB(z80cpu, z80cpu.regHL, b);
        Z80CPA(z80cpu, b);
        Inc(z80cpu.regHL); Z80SetBC(z80cpu, Z80GetBC(z80cpu)-1);
        z80cpu.flagC := c;
        c := Z80GetBC(z80cpu) <> 0; z80cpu.flagPV := c;
        if z80cpu.flagPV and not z80cpu.flagZ then
        begin
          Dec(z80cpu.regPC, 2);
          Inc(z80cpu.tStates, 5);
        end;
        //!FUSE!
        b := z80cpu.regA-b; if z80cpu.flagH then Dec(b);
        z80cpu.flag3 := (b and Z80FLAG3) <> 0;
        z80cpu.flag5 := (b and $02) <> 0;
      end;
    178: //INIR
      begin
        Inc(z80cpu.tStates);
        Inc(z80cpu.tStates, 4);//PRD(4)
        b := Z80PortIn(z80cpu, Z80GetBC(z80cpu));
        Inc(z80cpu.tStates, 3);//MWR(3)
        Z80PokeB(z80cpu, z80cpu.regHL, b);
        z80cpu.regB := z80cpu.regB-1; Inc(z80cpu.regHL);
        z80cpu.flagZ := true;
        z80cpu.flagN := true;
        if z80cpu.regB <> 0 then
        begin
          Dec(z80cpu.regPC, 2);
          Inc(z80cpu.tStates, 5);
        end;
        //!FUSE!
        z80cpu.flagN := (b and $80) <> 0;
        b := z80cpu.regB;
        z80cpu.flagS := (b and $80) <> 0;
        z80cpu.flagZ := (b = 0);
        z80cpu.flag3 := (b and Z80FLAG3) <> 0;
        z80cpu.flag5 := (b and Z80FLAG5) <> 0;
        z80cpu.flagPV := false;
        z80cpu.flagH := false;
        z80cpu.flagC := false;
      end;
    179: //OTIR
      begin
        Inc(z80cpu.tStates);
        z80cpu.regB := z80cpu.regB-1;
        Inc(z80cpu.tStates, 3);//MRD(3)
        b := Z80PeekB(z80cpu, z80cpu.regHL);
        Inc(z80cpu.tStates, 4);//PWR(4)
        Z80PortOut(z80cpu, Z80GetBC(z80cpu), b);
        Inc(z80cpu.regHL);
        z80cpu.flagZ := true;
        z80cpu.flagN := true;
        if z80cpu.regB <> 0 then
        begin
          Dec(z80cpu.regPC, 2);
          Inc(z80cpu.tStates, 5);
        end;
        //!FUSE!
        z80cpu.flagN := (b and $80) <> 0;
        b := z80cpu.regB;
        z80cpu.flagS := (b and $80) <> 0;
        z80cpu.flagZ := (b = 0);
        z80cpu.flag3 := (b and Z80FLAG3) <> 0;
        z80cpu.flag5 := (b and Z80FLAG5) <> 0;
        z80cpu.flagPV := false;
        z80cpu.flagH := false;
        z80cpu.flagC := false;
      end;
    184: //LDDR
      begin
        count := Z80GetBC(z80cpu); dest := z80cpu.regDE; from := z80cpu.regHL;
        Inc(z80cpu.tStates, 3);//MRD(3)
        b := Z80PeekB(z80cpu, from);
        Inc(z80cpu.tStates, 5);//MWR(5)
        Z80PokeB(z80cpu, dest, b);
        Dec(from); Dec(dest); Dec(count);
        z80cpu.flagH := false;
        z80cpu.flagN := false;
        z80cpu.flagPV := (count <> 0);
        if count <> 0 then
        begin
          Dec(z80cpu.regPC, 2);
          Inc(z80cpu.tStates, 5); //IOP(5)
        end;
        z80cpu.regDE := dest; z80cpu.regHL := from; Z80SetBC(z80cpu, count);
        //!FUSE!
        Inc(b, z80cpu.regA);
        z80cpu.flag3 := (b and Z80FLAG3) <> 0;
        z80cpu.flag5 := (b and $02) <> 0;
      end;
    185: //CPDR
      begin
        c := z80cpu.flagC;
        Inc(z80cpu.tStates, 3);//MRD(3)
        b := Z80PeekB(z80cpu, z80cpu.regHL);
        Inc(z80cpu.tStates, 5);//MWR(5)
        Z80PokeB(z80cpu, z80cpu.regHL, b);
        Z80CPA(z80cpu, b);
        Dec(z80cpu.regHL); Z80SetBC(z80cpu, Z80GetBC(z80cpu)-1);
        z80cpu.flagC := c;
        c := Z80GetBC(z80cpu) <> 0; z80cpu.flagPV := c;
        if z80cpu.flagPV and not z80cpu.flagZ then
        begin
          Dec(z80cpu.regPC, 2);
          Inc(z80cpu.tStates, 5);
        end;
        //!FUSE!
        b := z80cpu.regA-b; if z80cpu.flagH then Dec(b);
        z80cpu.flag3 := (b and Z80FLAG3) <> 0;
        z80cpu.flag5 := (b and $02) <> 0;
      end;
    186: //INDR
      begin
        Inc(z80cpu.tStates);
        Inc(z80cpu.tStates, 4);//PRD(4)
        b := Z80PortIn(z80cpu, Z80GetBC(z80cpu));
        Inc(z80cpu.tStates, 3);//MWR(3)
        Z80PokeB(z80cpu, z80cpu.regHL, b);
        z80cpu.regB := z80cpu.regB-1; Dec(z80cpu.regHL);
        z80cpu.flagZ := true;
        z80cpu.flagN := true;
        if z80cpu.regB <> 0 then
        begin
          Dec(z80cpu.regPC, 2);
          Inc(z80cpu.tStates, 5);
        end;
        //!FUSE!
        z80cpu.flagN := (b and $80) <> 0;
        b := z80cpu.regB;
        z80cpu.flagS := (b and $80) <> 0;
        z80cpu.flagZ := (b = 0);
        z80cpu.flag3 := (b and Z80FLAG3) <> 0;
        z80cpu.flag5 := (b and Z80FLAG5) <> 0;
        z80cpu.flagPV := false;
        z80cpu.flagH := false;
        z80cpu.flagC := false;
      end;
    187: //OTDR
      begin
        Inc(z80cpu.tStates);
        z80cpu.regB := z80cpu.regB-1;
        Inc(z80cpu.tStates, 3);//MRD(3)
        b := Z80PeekB(z80cpu, z80cpu.regHL);
        Inc(z80cpu.tStates, 4);//PWR(4)
        Z80PortOut(z80cpu, Z80GetBC(z80cpu), b);
        Dec(z80cpu.regHL);
        z80cpu.flagZ := true;
        z80cpu.flagN := true;
        if z80cpu.regB <> 0 then
        begin
          Dec(z80cpu.regPC, 2);
          Inc(z80cpu.tStates, 5);
        end;
        //!FUSE!
        z80cpu.flagN := (b and $80) <> 0;
        b := z80cpu.regB;
        z80cpu.flagS := (b and $80) <> 0;
        z80cpu.flagZ := (b = 0);
        z80cpu.flag3 := (b and Z80FLAG3) <> 0;
        z80cpu.flag5 := (b and Z80FLAG5) <> 0;
        z80cpu.flagPV := false;
        z80cpu.flagH := false;
        z80cpu.flagC := false;
      end;
  end;
end;

procedure Z80ExecuteID (z80cpu: PZ80CPU; xxx: Byte);
var
  lTemp, op: Word;
  b: Byte;
begin
  case xxx of
    9: //ADD ID,BC
      begin
        z80cpu.regID := Z80Add16(z80cpu, z80cpu.regID, Z80GetBC(z80cpu));
        Inc(z80cpu.tStates, 7);//IOP(4),IOP(3)
      end;
    25: //ADD ID,DE
      begin
        z80cpu.regID := Z80Add16(z80cpu, z80cpu.regID, z80cpu.regDE);
        Inc(z80cpu.tStates, 7);//IOP(4),IOP(3)
      end;
    41: //ADD ID,ID
      begin
        lTemp := z80cpu.regID;
        z80cpu.regID := Z80Add16(z80cpu, lTemp, lTemp);
        Inc(z80cpu.tStates, 7);//IOP(4),IOP(3)
      end;
    57: //ADD ID,SP
      begin
        z80cpu.regID := Z80Add16(z80cpu, z80cpu.regID, z80cpu.regSP);
        Inc(z80cpu.tStates, 7);//IOP(4),IOP(3)
      end;
    33: //LD ID,nn
      z80cpu.regID := LoadPCWord6T(z80cpu);
    34: //LD (nn),ID
      PokeWord6T(z80cpu, LoadPCWord6T(z80cpu), z80cpu.regID);
    42: //LD ID,(nn)
      z80cpu.regID := PeekWord6T(z80cpu, LoadPCWord6T(z80cpu));
    35: //INC ID
      begin
        Inc(z80cpu.tStates, 2);//OCR(6)
        Inc(z80cpu.regID);
      end;
    43: //DEC ID
      begin
        Inc(z80cpu.tStates, 2);//OCR(6)
        Dec(z80cpu.regID);
      end;
    36: //INC IDH
      begin
        //!FUSE!Inc(z80cpu.tStates);//OCR(5)
        Z80SetIDH(z80cpu, Z80Inc8(z80cpu, Z80GetIDH(z80cpu)));
      end;
    44: //INC IDL
      begin
        //!FUSE!Inc(z80cpu.tStates);//OCR(5)
        Z80SetIDL(z80cpu, Z80Inc8(z80cpu, Z80GetIDL(z80cpu)));
      end;
    52: //INC (ID+d)
      begin
        lTemp := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 5); //IOP(5)
        Inc(z80cpu.tStates, 4); //MRD(4)
        op := Z80PeekB(z80cpu, lTemp);
        Inc(z80cpu.tStates, 3); //MWR(3)
        Z80PokeB(z80cpu, lTemp, Z80Inc8(z80cpu, op));
      end;
    37: //DEC IDH
      begin
        //!FUSE!Inc(z80cpu.tStates);//OCR(5)
        Z80SetIDH(z80cpu, Z80Dec8(z80cpu, Z80GetIDH(z80cpu)));
      end;
    45: //DEC IDL
      begin
        //!FUSE!Inc(z80cpu.tStates);//OCR(5)
        Z80SetIDL(z80cpu, Z80Dec8(z80cpu, Z80GetIDL(z80cpu)));
      end;
    53: //DEC (ID+d)
      begin
        lTemp := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 5); //IOP(5)
        Inc(z80cpu.tStates, 4); //MRD(4)
        op := Z80PeekB(z80cpu, lTemp);
        Inc(z80cpu.tStates, 3); //MWR(3)
        Z80PokeB(z80cpu, lTemp, Z80Dec8(z80cpu, op));
      end;
    38: //LD IDH,n
      Z80SetIDH(z80cpu, LoadPCByte3T(z80cpu));
    46: //LD IDL,n
      Z80SetIDL(z80cpu, LoadPCByte3T(z80cpu));
    54: //LD (ID+d),n
      begin
        lTemp := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 2); //IOP(2)
        b := LoadPCByte3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MWR(3)
        Z80PokeB(z80cpu, lTemp, b);
      end;
    68: //LD B,IDH
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        z80cpu.regB := Z80GetIDH(z80cpu);
      end;
    69: //LD B,IDL
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        z80cpu.regB := Z80GetIDL(z80cpu);
      end;
    70: //LD B,(ID+d)
      begin
        Inc(z80cpu.tStates, 5); //IOP(5)
        op := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MRD(3)
        z80cpu.regB := Z80PeekB(z80cpu, op);
      end;
    76: //LD C,IDH
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        z80cpu.regC := Z80GetIDH(z80cpu);
      end;
    77: //LD C,IDL
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        z80cpu.regC := Z80GetIDL(z80cpu);
      end;
    78: //LD C,(ID+d)
      begin
        Inc(z80cpu.tStates, 5); //IOP(5)
        op := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MRD(3)
        z80cpu.regC := Z80PeekB(z80cpu, op);
      end;
    84: //LD D,IDH
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80SetD(z80cpu, Z80GetIDH(z80cpu));
      end;
    85: //LD D,IDL
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80SetD(z80cpu, Z80GetIDL(z80cpu));
      end;
    86: //LD D,(ID+d)
      begin
        Inc(z80cpu.tStates, 5); //IOP(5)
        op := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MRD(3)
        Z80SetD(z80cpu, Z80PeekB(z80cpu, op));
      end;
    92: //LD E,IDH
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80SetE(z80cpu, Z80GetIDH(z80cpu));
      end;
    93: //LD E,IDL
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80SetE(z80cpu, Z80GetIDL(z80cpu));
      end;
    94: //LD E,(ID+d)
      begin
        Inc(z80cpu.tStates, 5); //IOP(5)
        op := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MRD(3)
        Z80SetE(z80cpu, Z80PeekB(z80cpu, op));
      end;
    96: //LD IDH,B
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80SetIDH(z80cpu, z80cpu.regB);
      end;
    97: //LD IDH,C
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80SetIDH(z80cpu, z80cpu.regC);
      end;
    98: //LD IDH,D
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80SetIDH(z80cpu, (z80cpu.regDE shr 8));
      end;
    99: //LD IDH,E
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80SetIDH(z80cpu, (z80cpu.regDE and $FF));
      end;
    100:;//LD IDH,IDH
      //!FUSE!Inc(z80cpu.tStates); //IOP(1)
    101: //LD IDH,IDL
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80SetIDH(z80cpu, Z80GetIDL(z80cpu));
      end;
    102: //LD H,(ID+d)
      begin
        Inc(z80cpu.tStates, 5); //IOP(5)
        op := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MRD(3)
        Z80SetH(z80cpu, Z80PeekB(z80cpu, op));
      end;
    103: //LD IDH,A
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80SetIDH(z80cpu, z80cpu.regA);
      end;
    104: //LD IDL,B
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80SetIDL(z80cpu, z80cpu.regB);
      end;
    105: //LD IDL,C
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80SetIDL(z80cpu, z80cpu.regC);
      end;
    106: //LD IDL,D
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80SetIDL(z80cpu, (z80cpu.regDE shr 8));
      end;
    107: //LD IDL,E
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80SetIDL(z80cpu, (z80cpu.regDE and $FF));
      end;
    108: //LD IDL,IDH
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80SetIDL(z80cpu, Z80GetIDH(z80cpu));
      end;
    109:;//LD IDL,IDL
      //!FUSE!Inc(z80cpu.tStates); //IOP(1)
    110: //LD L,(ID+d)
      begin
        Inc(z80cpu.tStates, 5); //IOP(5)
        op := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MRD(3)
        Z80SetL(z80cpu, Z80PeekB(z80cpu, op));
      end;
    111: //LD IDL,A
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80SetIDL(z80cpu, z80cpu.regA);
      end;
    112: //LD (ID+d),B
      begin
        Inc(z80cpu.tStates, 5); //IOP(5)
        op := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MWR(3)
        Z80PokeB(z80cpu, op, z80cpu.regB);
      end;
    113: //LD (ID+d),C
      begin
        Inc(z80cpu.tStates, 5); //IOP(5)
        op := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MWR(3)
        Z80PokeB(z80cpu, op, z80cpu.regC);
      end;
    114: //LD (ID+d),D
      begin
        Inc(z80cpu.tStates, 5); //IOP(5)
        op := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MWR(3)
        Z80PokeB(z80cpu, op, (z80cpu.regDE shr 8));
      end;
    115: //LD (ID+d),E
      begin
        Inc(z80cpu.tStates, 5); //IOP(5)
        op := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MWR(3)
        Z80PokeB(z80cpu, op, (z80cpu.regDE and $FF));
      end;
    116: //LD (ID+d),H
      begin
        Inc(z80cpu.tStates, 5); //IOP(5)
        op := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MWR(3)
        Z80PokeB(z80cpu, op, (z80cpu.regHL shr 8));
      end;
    117: //LD (ID+d),L
      begin
        Inc(z80cpu.tStates, 5); //IOP(5)
        op := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MWR(3)
        Z80PokeB(z80cpu, op, (z80cpu.regHL and $FF));
      end;
    119: //LD (ID+d),A
      begin
        Inc(z80cpu.tStates, 5); //IOP(5)
        op := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MWR(3)
        Z80PokeB(z80cpu, op, z80cpu.regA);
      end;
    124: //LD A,IDH
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        z80cpu.regA := Z80GetIDH(z80cpu);
      end;
    125: //LD A,IDL
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        z80cpu.regA := Z80GetIDL(z80cpu);
      end;
    126: //LD A,(ID+d)
      begin
        Inc(z80cpu.tStates, 5); //IOP(5)
        op := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MRD(3)
        z80cpu.regA := Z80PeekB(z80cpu, op);
      end;
    132: //ADD A,IDH
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80AddA(z80cpu, Z80GetIDH(z80cpu));
      end;
    133: //ADD A,IDL
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80AddA(z80cpu, Z80GetIDL(z80cpu));
      end;
    134: //ADD A,(ID+d)
      begin
        Inc(z80cpu.tStates, 5); //IOP(5)
        op := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MRD(3)
        Z80AddA(z80cpu, Z80PeekB(z80cpu, op));
      end;
    140: //ADC A,IDH
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80AdcA(z80cpu, Z80GetIDH(z80cpu));
      end;
    141: //ADC A,IDL
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80AdcA(z80cpu, Z80GetIDL(z80cpu));
      end;
    142: //ADC A,(ID+d)
      begin
        Inc(z80cpu.tStates, 5); //IOP(5)
        op := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MRD(3)
        Z80AdcA(z80cpu, Z80PeekB(z80cpu, op));
      end;
    148: //SUB IDH
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80SubA(z80cpu, Z80GetIDH(z80cpu));
      end;
    149: //SUB IDL
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80SubA(z80cpu, Z80GetIDL(z80cpu));
      end;
    150: //SUB (ID+d)
      begin
        Inc(z80cpu.tStates, 5); //IOP(5)
        op := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MRD(3)
        Z80SubA(z80cpu, Z80PeekB(z80cpu, op));
      end;
    156: //SBC A,IDH
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80SbcA(z80cpu, Z80GetIDH(z80cpu));
      end;
    157: //SBC A,IDL
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80SbcA(z80cpu, Z80GetIDL(z80cpu));
      end;
    158: //SBC A,(ID+d)
      begin
        Inc(z80cpu.tStates, 5); //IOP(5)
        op := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MRD(3)
        Z80SbcA(z80cpu, Z80PeekB(z80cpu, op));
      end;
    164: //AND IDH
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80AndA(z80cpu, Z80GetIDH(z80cpu));
      end;
    165: //AND IDL
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80AndA(z80cpu, Z80GetIDL(z80cpu));
      end;
    166: //AND (ID+d)
      begin
        Inc(z80cpu.tStates, 5); //IOP(5)
        op := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MRD(3)
        Z80AndA(z80cpu, Z80PeekB(z80cpu, op));
      end;
    172: //XOR IDH
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80XorA(z80cpu, Z80GetIDH(z80cpu));
      end;
    173: //XOR IDL
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80XorA(z80cpu, Z80GetIDL(z80cpu));
      end;
    174: //XOR (ID+d) (???)
      begin
        Inc(z80cpu.tStates, 5); //IOP(5)
        op := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MRD(3)
        Z80XorA(z80cpu, Z80PeekB(z80cpu, op));
      end;
    180: //OR IDH
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80OrA(z80cpu, Z80GetIDH(z80cpu));
      end;
    181: //OR IDL
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80OrA(z80cpu, Z80GetIDL(z80cpu));
      end;
    182: //OR (ID+d)
      begin
        Inc(z80cpu.tStates, 5); //IOP(5)
        op := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MRD(3)
        Z80OrA(z80cpu, Z80PeekB(z80cpu, op));
      end;
    188: //CP IDH
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80CPA(z80cpu, Z80GetIDH(z80cpu));
      end;
    189: //CP IDL
      begin
        //!FUSE!Inc(z80cpu.tStates); //IOP(1)
        Z80CPA(z80cpu, Z80GetIDL(z80cpu));
      end;
    190: //CP (ID+d)
      begin
        Inc(z80cpu.tStates, 5); //IOP(5)
        op := Z80IDD3T(z80cpu);
        Inc(z80cpu.tStates, 3); //MRD(3)
        Z80CPA(z80cpu, Z80PeekB(z80cpu, op));
      end;
    203: //prefix CB
      begin
        lTemp := Z80IDD3T(z80cpu);
        op := Z80PeekB(z80cpu, z80cpu.regPC); Inc(z80cpu.regPC);
        Inc(z80cpu.tStates, 3); //???checkme!
        Z80ExecuteIDCB(z80cpu, op, lTemp);
        //K8:??? additional MWR(3) here? emulate by MRD(3)
        Inc(z80cpu.tStates, 3); //MRD(3)
        Z80PeekB(z80cpu, lTemp); //tst -- comment me!
      end;
    225: //POP ID
      z80cpu.regID := Z80Pop6T(z80cpu);
    227: //EX (SP),ID
      begin
        lTemp := z80cpu.regID;
        z80cpu.regID := PeekWord6T(z80cpu, z80cpu.regSP);
        Inc(z80cpu.tStates); //SRH(4)
        Inc(z80cpu.tStates, 3); //SWH(3)
        Z80PokeB(z80cpu, z80cpu.regSP+1, lTemp shr 8);
        Inc(z80cpu.tStates, 5); //SWL(5)
        Z80PokeB(z80cpu, z80cpu.regSP, lTemp and $FF);
      end;
    229: //PUSH ID
      begin
        Inc(z80cpu.tStates); //IOP(1)
        Z80Push6T(z80cpu, z80cpu.regID);
      end;
    233: //JP ID
      z80cpu.regPC := z80cpu.regID;
    249: //LD SP,ID
      begin
        Inc(z80cpu.tStates, 2); //IOP(2)
        z80cpu.regSP := z80cpu.regID;
      end;
  end;
end;


const
  notIDOpSet: set of Byte = [
    0..8, 10..24, 26..32, 39, 40, 47..51, 55, 56,
    58..67, 71..75, 79..83, 87..91, 95, 118,
    120..123, 127..131, 135..139, 143..147,
    151..155, 159..163, 167..171, 175..179,
    183..187, 191..202, 204..224, 226, 228,
    230..232, 234..248, 250..255];

procedure Z80Execute (z80cpu: PZ80CPU; stoptStates: Integer);
label again;
var
  xxx, d, lTemp, tmpW: Word;
  tmpB: Byte;
  prIX: Boolean;
  f: Integer;
begin
  z80cpu.breaked := false;
  if z80cpu.halted then exit;
  while (z80cpu.tStates < stoptStates) and not z80cpu.breaked do
  begin
again:
    z80cpu.prevPC := z80cpu.orgPC;
    z80cpu.orgPC := z80cpu.regPC;
    // m1_cycle
    xxx := LoadPCByteExec4T(z80cpu);//OCR(4)
    if z80cpu.breaked then exit;

    Inc(z80cpu.regRTemp);
    {$IFDEF USE_ONEMOREREAD}
    if z80cpu.oneMoreRead then
    begin
      d := z80cpu.regI;
      if (d >= $40) and (d <= $7F) then Z80PeekB(z80cpu, d*256);
    end;
    {$ENDIF}

    {$IFDEF USE_TRAPS}
      {$IFNDEF ONLY_ED_TRAPS}
      if Assigned(Z80OpTraps[xxx]) then
      begin
        if Z80OpTraps[xxx](z80cpu) then continue;
        // when this is called, opcode already fetched, PC & R incremented by 1,
        // t-states incremented by 4 (+possible contention)
        // return FALSE to continue as usual
        // return TRUE to fetch & run new instruction
        // set <breaked> to break immediately (res=TRUE) or after interpretation
        // of the current opcode (res=FALSE). no state unrolling will be made
        // on immediate break
      end;
      {$ENDIF}
    {$ENDIF}
    case xxx of
      0:;//NOP
      8: //EX AF,AF'
        begin
          if (z80cpu.regPC <> $04D1) or not Assigned(Z80SaveTrap) or
             not Z80SaveTrap(z80cpu) then Z80ExAFAF(z80cpu);
        end;
      16: //DJNZ dis
        begin
          Inc(z80cpu.tStates); //OCR(5)
          d := LoadPCByte3T(z80cpu); //ORD(3)
          lTemp := (z80cpu.regB-1) and $FF; z80cpu.regB := lTemp;
          if lTemp <> 0 then
          begin
            if (d and 128) = 128 then d := -(256-d);
            //IOP(5)
            if Assigned(Z80Contention) then
              for f := 0 to 4 do begin Z80Contention(z80cpu, z80cpu.regPC-1); Inc(z80cpu.tStates); end;
            z80cpu.regPC := (z80cpu.regPC+d) and $FFFF;
            if not Assigned(Z80Contention) then Inc(z80cpu.tStates, 5); //IOP(5)
          end;
        end;
      24: //JR dis
        begin
          d := LoadPCByte3T(z80cpu); //ORD(3)
          if (d and 128) = 128 then d := -(256-d);
          //IOP(5)
          if Assigned(Z80Contention) then
            for f := 0 to 4 do begin Z80Contention(z80cpu, z80cpu.regPC-1); Inc(z80cpu.tStates); end;
          z80cpu.regPC := (z80cpu.regPC+d) and $FFFF;
          if not Assigned(Z80Contention) then Inc(z80cpu.tStates, 5); //IOP(5)
        end;
      32: //JR NZ,dis
        begin
          d := LoadPCByte3T(z80cpu); //ORD(3)
          if not z80cpu.flagZ then
          begin
            if (d and 128) = 128 then d := -(256-d);
            //IOP(5)
            if Assigned(Z80Contention) then
              for f := 0 to 4 do begin Z80Contention(z80cpu, z80cpu.regPC-1); Inc(z80cpu.tStates); end;
            z80cpu.regPC := ((z80cpu.regPC+d) and $FFFF);
            if not Assigned(Z80Contention) then Inc(z80cpu.tStates, 5); //IOP(5)
          end;
        end;
      40: //JR Z, dis
        begin
          d := LoadPCByte3T(z80cpu); //ORD(3)
          if z80cpu.flagZ then
          begin
            if (d and 128) = 128 then d := -(256-d);
            //IOP(5)
            if Assigned(Z80Contention) then
              for f := 0 to 4 do begin Z80Contention(z80cpu, z80cpu.regPC-1); Inc(z80cpu.tStates); end;
            z80cpu.regPC := ((z80cpu.regPC+d) and $FFFF);
            if not Assigned(Z80Contention) then Inc(z80cpu.tStates, 5); //IOP(5)
          end;
        end;
      48: //JR NC, dis
        begin
          d := LoadPCByte3T(z80cpu); //ORD(3)
          if not z80cpu.flagC then
          begin
            if (d and 128) = 128 then d := -(256-d);
            //IOP(5)
            if Assigned(Z80Contention) then
              for f := 0 to 4 do begin Z80Contention(z80cpu, z80cpu.regPC-1); Inc(z80cpu.tStates); end;
            z80cpu.regPC := ((z80cpu.regPC+d) and $FFFF);
            if not Assigned(Z80Contention) then Inc(z80cpu.tStates, 5); //IOP(5)
          end;
        end;
      56: //JR C, dis
        begin
          d := LoadPCByte3T(z80cpu); //ORD(3)
          if z80cpu.flagC then
          begin
            if (d and 128) = 128 then d := -(256-d);
            //IOP(5)
            if Assigned(Z80Contention) then
              for f := 0 to 4 do begin Z80Contention(z80cpu, z80cpu.regPC-1); Inc(z80cpu.tStates); end;
            z80cpu.regPC := ((z80cpu.regPC+d) and $FFFF);
            if not Assigned(Z80Contention) then Inc(z80cpu.tStates, 5); //IOP(5)
          end;
        end;
      1: //LD BC,nn
        Z80SetBC(z80cpu, LoadPCWord6T(z80cpu));//ORL(3),ORH(3)
      9: //ADD HL,BC
        begin
          z80cpu.regHL := Z80Add16(z80cpu, z80cpu.regHL, Z80GetBC(z80cpu));
          Inc(z80cpu.tStates, 7);//IOP(4),IOP(3)
        end;
      17: //LD DE,nn
        z80cpu.regDE := LoadPCWord6T(z80cpu);//ORL(3),ORH(3)
      25: //ADD HL,DE
        begin
          z80cpu.regHL := Z80Add16(z80cpu, z80cpu.regHL, z80cpu.regDE);
          Inc(z80cpu.tStates, 7);//IOP(4),IOP(3)
        end;
      33: //LD HL,nn
        z80cpu.regHL := LoadPCWord6T(z80cpu);//ORL(3),ORH(3)
      41: //ADD HL,HL
        begin
          z80cpu.regHL := Z80Add16(z80cpu, z80cpu.regHL, z80cpu.regHL);
          Inc(z80cpu.tStates, 7);//IOP(4),IOP(3)
        end;
      49: //LD SP,nn
        z80cpu.regSP := LoadPCWord6T(z80cpu);//ORL(3),ORH(3)
      57: //ADD HL,SP
        begin
          z80cpu.regHL := Z80Add16(z80cpu, z80cpu.regHL, z80cpu.regSP);
          Inc(z80cpu.tStates, 7);//IOP(4),IOP(3)
        end;
      2: //LD (BC),A
        begin
          Inc(z80cpu.tStates, 3);//MWR(3)
          Z80PokeB(z80cpu, Z80GetBC(z80cpu), z80cpu.regA);
        end;
      10: //LD A,(BC)
        begin
          Inc(z80cpu.tStates, 3);//MRD(3)
          z80cpu.regA := Z80PeekB(z80cpu, Z80GetBC(z80cpu));
        end;
      18: //LD (DE),A
        begin
          Inc(z80cpu.tStates, 3);//MWR(3)
          Z80PokeB(z80cpu, z80cpu.regDE, z80cpu.regA);
        end;
      26: //LD A,(DE)
        begin
          Inc(z80cpu.tStates, 3);//MRD(3)
          z80cpu.regA := Z80PeekB(z80cpu, z80cpu.regDE);
        end;
      34: //LD (nn),HL
        begin
          lTemp := LoadPCWord6T(z80cpu);//ORL(3),ORH(3)
          PokeWord6T(z80cpu, lTemp, z80cpu.regHL);
        end;
      42: //LD HL,(nn)
        begin
          lTemp := LoadPCWord6T(z80cpu);//ORL(3),ORH(3)
          z80cpu.regHL := PeekWord6T(z80cpu, lTemp);
        end;
      50: //LD (nn),A
        begin
          tmpW := LoadPCWord6T(z80cpu);
          Inc(z80cpu.tStates, 3);//MWR(3)
          Z80PokeB(z80cpu, tmpW, z80cpu.regA);//ORL(3),ORH(3)
        end;
      58: //LD A,(nn)
        begin
          tmpW := LoadPCWord6T(z80cpu);
          Inc(z80cpu.tStates, 3);//MRD(3)
          z80cpu.regA := Z80PeekB(z80cpu, tmpW);//ORL(3),ORH(3)
        end;
      3: //INC BC
        begin
          Inc(z80cpu.tStates, 2);//OCR(6)
          Z80SetBC(z80cpu, Z80GetBC(z80cpu)+1);
        end;
      11: //DEC BC
        begin
          Inc(z80cpu.tStates, 2);//OCR(6)
          Z80SetBC(z80cpu, Z80GetBC(z80cpu)-1);
        end;
      19: //INC DE
        begin
          Inc(z80cpu.tStates, 2);//OCR(6)
          Inc(z80cpu.regDE);
        end;
      27: //DEC DE
        begin
          Inc(z80cpu.tStates, 2);//OCR(6)
          Dec(z80cpu.regDE);
        end;
      35: //INC HL
        begin
          Inc(z80cpu.tStates, 2);//OCR(6)
          Inc(z80cpu.regHL);
        end;
      43: //DEC HL
        begin
          Inc(z80cpu.tStates, 2);//OCR(6)
          Dec(z80cpu.regHL);
        end;
      51: //INC SP
        begin
          Inc(z80cpu.tStates, 2);//OCR(6)
          Inc(z80cpu.regSP);
        end;
      59: //DEC SP
        begin
          Inc(z80cpu.tStates, 2);//OCR(6)
          Dec(z80cpu.regSP);
        end;
      4: //INC B
        z80cpu.regB := Z80Inc8(z80cpu, z80cpu.regB);
      12: //INC C
        z80cpu.regC := Z80Inc8(z80cpu, z80cpu.regC);
      20: //INC D
        Z80SetD(z80cpu, Z80Inc8(z80cpu, (z80cpu.regDE shr 8)));
      28: //INC E
        Z80SetE(z80cpu, Z80Inc8(z80cpu, (z80cpu.regDE and $FF)));
      36: //INC H
        Z80SetH(z80cpu, Z80Inc8(z80cpu, (z80cpu.regHL shr 8)));
      44: //INC L
        Z80SetL(z80cpu, Z80Inc8(z80cpu, (z80cpu.regHL and $FF)));
      52: //INC (HL)
        begin
          Inc(z80cpu.tStates); //OCR(5)
          Inc(z80cpu.tStates, 3);//MRD(3)
          lTemp := Z80PeekB(z80cpu, z80cpu.regHL);
          Inc(z80cpu.tStates, 3);//MWR(3)
          Z80PokeB(z80cpu, z80cpu.regHL, Z80Inc8(z80cpu, lTemp));
        end;
      60: //INC A
        z80cpu.regA := Z80Inc8(z80cpu, z80cpu.regA);
      5: //DEC B
        z80cpu.regB := Z80Dec8(z80cpu, z80cpu.regB);
      13: //DEC C
        z80cpu.regC := Z80Dec8(z80cpu, z80cpu.regC);
      21: //DEC D
        Z80SetD(z80cpu, Z80Dec8(z80cpu, (z80cpu.regDE shr 8)));
      29: //DEC E
        Z80SetE(z80cpu, Z80Dec8(z80cpu, (z80cpu.regDE and $FF)));
      37: //DEC H
        Z80SetH(z80cpu, Z80Dec8(z80cpu, (z80cpu.regHL shr 8)));
      45: //DEC L
        Z80SetL(z80cpu, Z80Dec8(z80cpu, (z80cpu.regHL and $FF)));
      53: //DEC (HL)
        begin
          Inc(z80cpu.tStates); //OCR(5)
          Inc(z80cpu.tStates, 3);//MRD(3)
          lTemp := Z80PeekB(z80cpu, z80cpu.regHL);
          Inc(z80cpu.tStates, 3);//MWR(3)
          Z80PokeB(z80cpu, z80cpu.regHL, Z80Dec8(z80cpu, lTemp));
        end;
      61: //DEC A
        z80cpu.regA := Z80Dec8(z80cpu, z80cpu.regA);
      6: //LD B,n
        z80cpu.regB := LoadPCByte3T(z80cpu);
      14: //LD C,n
        z80cpu.regC := LoadPCByte3T(z80cpu);
      22: //LD D,n
        Z80SetD(z80cpu, LoadPCByte3T(z80cpu));
      30: //LD E,n
        Z80SetE(z80cpu, LoadPCByte3T(z80cpu));
      38: //LD H,n
        Z80SetH(z80cpu, LoadPCByte3T(z80cpu));
      46: //LD L,n
        Z80SetL(z80cpu, LoadPCByte3T(z80cpu));
      54: //LD (HL),n
        begin
          tmpB := LoadPCByte3T(z80cpu);
          Inc(z80cpu.tStates, 3); //MWR(3)
          Z80PokeB(z80cpu, z80cpu.regHL, tmpB);
        end;
      62: //LD A,n
        z80cpu.regA := LoadPCByte3T(z80cpu);
      7: //RLCA
        Z80RLCA(z80cpu);
      15: //RRCA
        Z80RRCA(z80cpu);
      23: //RLA
        Z80RLA(z80cpu);
      31: //RRA
        Z80RRA(z80cpu);
      39: //DAA
        Z80DAAA(z80cpu);
      47: //CPL
        Z80CPLA(z80cpu);
      55: //SCF
        Z80SCF(z80cpu);
      63: //CCF
        Z80CCF(z80cpu);
      64:;//LD B,B
      65: //LD B,C
        z80cpu.regB := z80cpu.regC;
      66: //LD B,D
        z80cpu.regB := (z80cpu.regDE shr 8);
      67: //LD B,E
        z80cpu.regB := (z80cpu.regDE and $FF);
      68: //LD B,H
        z80cpu.regB := (z80cpu.regHL shr 8);
      69: //LD B,L
        z80cpu.regB := (z80cpu.regHL and $FF);
      70: //LD B,(HL)
        begin
          Inc(z80cpu.tStates, 3); //MRD(3)
          z80cpu.regB := Z80PeekB(z80cpu, z80cpu.regHL);
        end;
      71: //LD B,A
        z80cpu.regB := z80cpu.regA;
      72: //LD C,B
        z80cpu.regC := z80cpu.regB;
      73:;//LD C,C
      74: //LD C,D
        z80cpu.regC := (z80cpu.regDE shr 8);
      75: //LD C,E
        z80cpu.regC := (z80cpu.regDE and $FF);
      76: //LD C,H
        z80cpu.regC := (z80cpu.regHL shr 8);
      77: //LD C,L
        z80cpu.regC := (z80cpu.regHL and $FF);
      78: //LD C,(HL)
        begin
          Inc(z80cpu.tStates, 3); //MRD(3)
          z80cpu.regC := Z80PeekB(z80cpu, z80cpu.regHL);
        end;
      79: //LD C,A
        z80cpu.regC := z80cpu.regA;
      80: //LD D,B
        Z80SetD(z80cpu, z80cpu.regB);
      81: //LD D,C
        Z80SetD(z80cpu, z80cpu.regC);
      82:;//LD D,D
      83: //LD D,E
        Z80SetD(z80cpu, (z80cpu.regDE and $FF));
      84: //LD D,H
        Z80SetD(z80cpu, (z80cpu.regHL shr 8));
      85: //LD D,L
        Z80SetD(z80cpu, (z80cpu.regHL and $FF));
      86: //LD D,(HL)
        begin
          Inc(z80cpu.tStates, 3); //MRD(3)
          Z80SetD(z80cpu, Z80PeekB(z80cpu, z80cpu.regHL));
        end;
      87: //LD D,A
        Z80SetD(z80cpu, z80cpu.regA);
      88: //LD E,B
        Z80SetE(z80cpu, z80cpu.regB);
      89: //LD E,C
        Z80SetE(z80cpu, z80cpu.regC);
      90: //LD E,D
        Z80SetE(z80cpu, (z80cpu.regDE shr 8));
      91:;//LD E,E
      92: //LD E,H
        Z80SetE(z80cpu, (z80cpu.regHL shr 8));
      93: //LD E,L
        Z80SetE(z80cpu, (z80cpu.regHL and $FF));
      94: //LD E,(HL)
        begin
          Inc(z80cpu.tStates, 3); //MRD(3)
          Z80SetE(z80cpu, Z80PeekB(z80cpu, z80cpu.regHL));
        end;
      95: //LD E,A
        Z80SetE(z80cpu, z80cpu.regA);
      96: //LD H,B
        Z80SetH(z80cpu, z80cpu.regB);
      97: //LD H,C
        Z80SetH(z80cpu, z80cpu.regC);
      98: //LD H,D
        Z80SetH(z80cpu, (z80cpu.regDE shr 8));
      99: //LD H,E
        Z80SetH(z80cpu, (z80cpu.regDE and $FF));
      100:;//LD H,H
      101: //LD H,L
        Z80SetH(z80cpu, (z80cpu.regHL and $FF));
      102: //LD H,(HL)
        begin
          Inc(z80cpu.tStates, 3); //MRD(3)
          Z80SetH(z80cpu, Z80PeekB(z80cpu, z80cpu.regHL));
        end;
      103: //LD H,A
        Z80SetH(z80cpu, z80cpu.regA);
      104: //LD L,B
        Z80SetL(z80cpu, z80cpu.regB);
      105: //LD L,C
        Z80SetL(z80cpu, z80cpu.regC);
      106: //LD L,D
        Z80SetL(z80cpu, (z80cpu.regDE shr 8));
      107: //LD L,E
        Z80SetL(z80cpu, (z80cpu.regDE and $FF));
      108: //LD L,H
        Z80SetL(z80cpu, (z80cpu.regHL shr 8));
      109:;//LD L,L
      110: //LD L,(HL)
        begin
          Inc(z80cpu.tStates, 3); //MRD(3)
          Z80SetL(z80cpu, Z80PeekB(z80cpu, z80cpu.regHL));
        end;
      111: //LD L,A
        Z80SetL(z80cpu, z80cpu.regA);
      112: //LD (HL),B
        begin
          Inc(z80cpu.tStates, 3); //MWR(3)
          Z80PokeB(z80cpu, z80cpu.regHL, z80cpu.regB);
        end;
      113: //LD (HL),C
        begin
          Inc(z80cpu.tStates, 3); //MWR(3)
          Z80PokeB(z80cpu, z80cpu.regHL, z80cpu.regC);
        end;
      114: //LD (HL),D
        begin
          Inc(z80cpu.tStates, 3); //MWR(3)
          Z80PokeB(z80cpu, z80cpu.regHL, (z80cpu.regDE shr 8));
        end;
      115: //LD (HL),E
        begin
          Inc(z80cpu.tStates, 3); //MWR(3)
          Z80PokeB(z80cpu, z80cpu.regHL, (z80cpu.regDE and $FF));
        end;
      116: //LD (HL),H
        begin
          Inc(z80cpu.tStates, 3); //MWR(3)
          Z80PokeB(z80cpu, z80cpu.regHL, (z80cpu.regHL shr 8));
        end;
      117: //LD (HL),L
        begin
          Inc(z80cpu.tStates, 3); //MWR(3)
          Z80PokeB(z80cpu, z80cpu.regHL, (z80cpu.regHL and $FF));
        end;
      118: //HALT
        begin
          z80cpu.halted := true;
          exit;
        end;
      119: //LD (HL),A
        begin
          Inc(z80cpu.tStates, 3); //MWR(3)
          Z80PokeB(z80cpu, z80cpu.regHL, z80cpu.regA);
        end;
      120: //LD A,B
        z80cpu.regA := z80cpu.regB;
      121: //LD A,C
        z80cpu.regA := z80cpu.regC;
      122: //LD A,D
        z80cpu.regA := (z80cpu.regDE shr 8);
      123: //LD A,E
        z80cpu.regA := (z80cpu.regDE and $FF);
      124: //LD A,H
        z80cpu.regA := (z80cpu.regHL shr 8);
      125: //LD A,L
        z80cpu.regA := (z80cpu.regHL and $FF);
      126: //LD A,(HL)
        begin
          Inc(z80cpu.tStates, 3); //MRD(3)
          z80cpu.regA := Z80PeekB(z80cpu, z80cpu.regHL);
        end;
      127:;//LD A,A
      128: //ADD A,B
        Z80AddA(z80cpu, z80cpu.regB);
      129: //ADD A,C
        Z80AddA(z80cpu, z80cpu.regC);
      130: //ADD A,D
        Z80AddA(z80cpu, (z80cpu.regDE shr 8));
      131: //ADD A,E
        Z80AddA(z80cpu, (z80cpu.regDE and $FF));
      132: //ADD A,H
        Z80AddA(z80cpu, (z80cpu.regHL shr 8));
      133: //ADD A,L
        Z80AddA(z80cpu, (z80cpu.regHL and $FF));
      134: //ADD A,(HL)
        begin
          Inc(z80cpu.tStates, 3); //MRD(3)
          Z80AddA(z80cpu, Z80PeekB(z80cpu, z80cpu.regHL));
        end;
      135: //ADD A,A
        Z80AddA(z80cpu, z80cpu.regA);
      136: //ADC A,B
        Z80AdcA(z80cpu, z80cpu.regB);
      137: //ADC A,C
        Z80AdcA(z80cpu, z80cpu.regC);
      138: //ADC A,D
        Z80AdcA(z80cpu, (z80cpu.regDE shr 8));
      139: //ADC A,E
        Z80AdcA(z80cpu, (z80cpu.regDE and $FF));
      140: //ADC A,H
        Z80AdcA(z80cpu, (z80cpu.regHL shr 8));
      141: //ADC A,L
        Z80AdcA(z80cpu, (z80cpu.regHL and $FF));
      142: //ADC A,(HL)
        begin
          Inc(z80cpu.tStates, 3); //MRD(3)
          Z80AdcA(z80cpu, Z80PeekB(z80cpu, z80cpu.regHL));
        end;
      143: //ADC A,A
        Z80AdcA(z80cpu, z80cpu.regA);
      144: //SUB B
        Z80SubA(z80cpu, z80cpu.regB);
      145: //SUB C
        Z80SubA(z80cpu, z80cpu.regC);
      146: //SUB D
        Z80SubA(z80cpu, (z80cpu.regDE shr 8));
      147: //SUB E
        Z80SubA(z80cpu, (z80cpu.regDE and $FF));
      148: //SUB H
        Z80SubA(z80cpu, (z80cpu.regHL shr 8));
      149: //SUB L
        Z80SubA(z80cpu, (z80cpu.regHL and $FF));
      150: //SUB (HL)
        begin
          Inc(z80cpu.tStates, 3); //MRD(3)
          Z80SubA(z80cpu, Z80PeekB(z80cpu, z80cpu.regHL));
        end;
      151: //SUB A
        Z80SubA(z80cpu, z80cpu.regA);
      152: //SBC A,B
        Z80SbcA(z80cpu, z80cpu.regB);
      153: //SBC A,C
        Z80SbcA(z80cpu, z80cpu.regC);
      154: //SBC A,D
        Z80SbcA(z80cpu, (z80cpu.regDE shr 8));
      155: //SBC A,E
        Z80SbcA(z80cpu, (z80cpu.regDE and $FF));
      156: //SBC A,H
        Z80SbcA(z80cpu, (z80cpu.regHL shr 8));
      157: //SBC A,L
        Z80SbcA(z80cpu, (z80cpu.regHL and $FF));
      158: //SBC A,(HL)
        begin
          Inc(z80cpu.tStates, 3); //MRD(3)
          Z80SbcA(z80cpu, Z80PeekB(z80cpu, z80cpu.regHL));
        end;
      159: //SBC A,A
        Z80SbcA(z80cpu, z80cpu.regA);
      160: //AND B
        Z80AndA(z80cpu, z80cpu.regB);
      161: //AND C
        Z80AndA(z80cpu, z80cpu.regC);
      162: //AND D
        Z80AndA(z80cpu, (z80cpu.regDE shr 8));
      163: //AND E
        Z80AndA(z80cpu, (z80cpu.regDE and $FF));
      164: //AND H
        Z80AndA(z80cpu, (z80cpu.regHL shr 8));
      165: //AND L
        Z80AndA(z80cpu, (z80cpu.regHL and $FF));
      166: //AND (HL)
        begin
          Inc(z80cpu.tStates, 3); //MRD(3)
          Z80AndA(z80cpu, Z80PeekB(z80cpu, z80cpu.regHL));
        end;
      167: //AND A
        Z80AndA(z80cpu, z80cpu.regA);
      168: //XOR B
        Z80XorA(z80cpu, z80cpu.regB);
      169: //XOR C
        Z80XorA(z80cpu, z80cpu.regC);
      170: //XOR D
        Z80XorA(z80cpu, (z80cpu.regDE shr 8));
      171: //XOR E
        Z80XorA(z80cpu, (z80cpu.regDE and $FF));
      172: //XOR H
        Z80XorA(z80cpu, (z80cpu.regHL shr 8));
      173: //XOR L
        Z80XorA(z80cpu, (z80cpu.regHL and $FF));
      174: //XOR (HL)
        begin
          Inc(z80cpu.tStates, 3); //MRD(3)
          Z80XorA(z80cpu, Z80PeekB(z80cpu, z80cpu.regHL));
        end;
      175: //XOR A
        begin
          z80cpu.regA := 0;
          z80cpu.flagS := false;
          z80cpu.flag3 := false;
          z80cpu.flag5 := false;
          z80cpu.flagH := false;
          z80cpu.flagPV := true;
          z80cpu.flagZ := true;
          z80cpu.flagN := false;
          z80cpu.flagC := false;
        end;
      176: //OR B
        Z80OrA(z80cpu, z80cpu.regB);
      177: //OR C
        Z80OrA(z80cpu, z80cpu.regC);
      178: //OR D
        Z80OrA(z80cpu, (z80cpu.regDE shr 8));
      179: //OR E
        Z80OrA(z80cpu, (z80cpu.regDE and $FF));
      180: //OR H
        Z80OrA(z80cpu, (z80cpu.regHL shr 8));
      181: //OR L
        Z80OrA(z80cpu, (z80cpu.regHL and $FF));
      182: //OR (HL)
        begin
          Inc(z80cpu.tStates, 3); //MRD(3)
          Z80OrA(z80cpu, Z80PeekB(z80cpu, z80cpu.regHL));
        end;
      183: //OR A
        Z80OrA(z80cpu, z80cpu.regA);
      184: //CP B
        Z80CPA(z80cpu, z80cpu.regB);
      185: //CP C
        Z80CPA(z80cpu, z80cpu.regC);
      186: //CP D
        Z80CPA(z80cpu, (z80cpu.regDE shr 8));
      187: //CP E
        Z80CPA(z80cpu, (z80cpu.regDE and $FF));
      188: //CP H
        Z80CPA(z80cpu, (z80cpu.regHL shr 8));
      189: //CP L
        Z80CPA(z80cpu, (z80cpu.regHL and $FF));
      190: //CP (HL)
        begin
          Inc(z80cpu.tStates, 3); //MRD(3)
          Z80CPA(z80cpu, Z80PeekB(z80cpu, z80cpu.regHL));
        end;
      191: //CP A
        Z80CPA(z80cpu, z80cpu.regA);
      192: //RET NZ
        begin
          Inc(z80cpu.tStates); //OCR(5)
          if not z80cpu.flagZ then z80cpu.regPC := Z80Pop6T(z80cpu);
        end;
      200: //RET Z
        begin
          Inc(z80cpu.tStates); //OCR(5)
          if z80cpu.flagZ then z80cpu.regPC := Z80Pop6T(z80cpu);
        end;
      208: //RET NC
        begin
          Inc(z80cpu.tStates); //OCR(5)
          if not z80cpu.flagC then z80cpu.regPC := Z80Pop6T(z80cpu);
        end;
      216: //RET C
        begin
          Inc(z80cpu.tStates); //OCR(5)
          if z80cpu.flagC then z80cpu.regPC := Z80Pop6T(z80cpu);
        end;
      224: //RET PO
        begin
          Inc(z80cpu.tStates); //OCR(5)
          if not z80cpu.flagPV then z80cpu.regPC := Z80Pop6T(z80cpu);
        end;
      232: //RET PE
        begin
          Inc(z80cpu.tStates); //OCR(5)
          if z80cpu.flagPV then z80cpu.regPC := Z80Pop6T(z80cpu);
        end;
      240: //RET P
        begin
          Inc(z80cpu.tStates); //OCR(5)
          if not z80cpu.flagS then z80cpu.regPC := Z80Pop6T(z80cpu);
        end;
      248: //RET M
        begin
          Inc(z80cpu.tStates); //OCR(5)
          if z80cpu.flagS then z80cpu.regPC := Z80Pop6T(z80cpu);
        end;
      193: //POP BC
        Z80SetBC(z80cpu, Z80Pop6T(z80cpu));
      201: //RET
        z80cpu.regPC := Z80Pop6T(z80cpu);
      209: //POP DE
        z80cpu.regDE := Z80Pop6T(z80cpu);
      217: //EXX
        Z80EXX(z80cpu);
      225: //POP HL
        z80cpu.regHL := Z80Pop6T(z80cpu);
      233: //JP HL
        z80cpu.regPC := z80cpu.regHL;
      241: //POP AF
        Z80SetAF(z80cpu, Z80Pop6T(z80cpu));
      249: //LD SP,HL
        begin
          Inc(z80cpu.tStates, 2); //OCR(6)
          z80cpu.regSP := z80cpu.regHL;
        end;
      194: //JP NZ,nn
        begin
          lTemp := LoadPCWord6T(z80cpu);
          if not z80cpu.flagZ then z80cpu.regPC := lTemp;
        end;
      202: //JP Z,nn
        begin
          lTemp := LoadPCWord6T(z80cpu);
          if z80cpu.flagZ then z80cpu.regPC := lTemp;
        end;
      210: //JP NC,nn
        begin
          lTemp := LoadPCWord6T(z80cpu);
          if not z80cpu.flagC then z80cpu.regPC := lTemp;
        end;
      218: //JP C,nn
        begin
          lTemp := LoadPCWord6T(z80cpu);
          if z80cpu.flagC then z80cpu.regPC := lTemp;
        end;
      226: //JP PO,nn
        begin
          lTemp := LoadPCWord6T(z80cpu);
          if not z80cpu.flagPV then z80cpu.regPC := lTemp;
        end;
      234: //JP PE,nn
        begin
          lTemp := LoadPCWord6T(z80cpu);
          if z80cpu.flagPV then z80cpu.regPC := lTemp;
        end;
      242: //JP P,nn
        begin
          lTemp := LoadPCWord6T(z80cpu);
          if not z80cpu.flagS then z80cpu.regPC := lTemp;
        end;
      250: //JP M,nn
        begin
          lTemp := LoadPCWord6T(z80cpu);
          if z80cpu.flagS then z80cpu.regPC := lTemp;
        end;
      195: //JP nn
        z80cpu.regPC := PeekWord6T(z80cpu, z80cpu.regPC);
      203: //prefix CB
        begin
          Z80ExecuteCB(z80cpu);
        end;
      211: //OUT (n),A
        begin
          tmpW := LoadPCByte3T(z80cpu)+z80cpu.regA*256;
          Inc(z80cpu.tStates, 4); //PWR(4)
          Z80PortOut(z80cpu, tmpW, z80cpu.regA);
        end;
      219: //IN A,(n)
        begin
          tmpW := LoadPCByte3T(z80cpu)+z80cpu.regA*256;
          Inc(z80cpu.tStates, 4); //PRD(4)
          z80cpu.regA := Z80PortIn(z80cpu, tmpW);
        end;
      227: //EX (SP),HL
        begin
          lTemp := z80cpu.regHL;
          z80cpu.regHL := PeekWord6T(z80cpu, z80cpu.regSP);
          Inc(z80cpu.tStates); //SRH(4)
          Inc(z80cpu.tStates, 3); //SWH(3)
          Z80PokeB(z80cpu, z80cpu.regSP+1, lTemp shr 8);
          Inc(z80cpu.tStates, 5); //SWL(5)
          Z80PokeB(z80cpu, z80cpu.regSP, lTemp and $FF);
        end;
      235: //EX DE,HL
        begin
          lTemp := z80cpu.regHL;
          z80cpu.regHL := z80cpu.regDE;
          z80cpu.regDE := lTemp;
        end;
      243: //DI
        begin
          z80cpu.iff1 := false;
          z80cpu.iff2 := false;
        end;
      251: //EI
        begin
          z80cpu.iff1 := true;
          z80cpu.iff2 := true;
          z80cpu.lastEItState := z80cpu.tStates;
        end;
      196: //CALL NZ,nn
        begin
          lTemp := LoadPCWord6T(z80cpu);
          if not z80cpu.flagZ then
          begin
            Inc(z80cpu.tStates); //ORH(4)
            Z80Push6T(z80cpu, z80cpu.regPC);
            z80cpu.regPC := lTemp;
          end;
        end;
      204: //CALL Z,nn
        begin
          lTemp := LoadPCWord6T(z80cpu);
          if z80cpu.flagZ then
          begin
            Inc(z80cpu.tStates); //ORH(4)
            Z80Push6T(z80cpu, z80cpu.regPC);
            z80cpu.regPC := lTemp;
          end;
        end;
      212: //CALL NC,nn
        begin
          lTemp := LoadPCWord6T(z80cpu);
          if not z80cpu.flagC then
          begin
            Inc(z80cpu.tStates); //ORH(4)
            Z80Push6T(z80cpu, z80cpu.regPC);
            z80cpu.regPC := lTemp;
          end;
        end;
      220: //CALL C,nn
        begin
          lTemp := LoadPCWord6T(z80cpu);
          if z80cpu.flagC then
          begin
            Inc(z80cpu.tStates); //ORH(4)
            Z80Push6T(z80cpu, z80cpu.regPC);
            z80cpu.regPC := lTemp;
          end;
        end;
      228: //CALL PO,nn
        begin
          lTemp := LoadPCWord6T(z80cpu);
          if not z80cpu.flagPV then
          begin
            Inc(z80cpu.tStates); //ORH(4)
            Z80Push6T(z80cpu, z80cpu.regPC);
            z80cpu.regPC := lTemp;
          end;
        end;
      236: //CALL PE,nn
        begin
          lTemp := LoadPCWord6T(z80cpu);
          if z80cpu.flagPV then
          begin
            Inc(z80cpu.tStates); //ORH(4)
            Z80Push6T(z80cpu, z80cpu.regPC);
            z80cpu.regPC := lTemp;
          end;
        end;
      244: //CALL P,nn
        begin
          lTemp := LoadPCWord6T(z80cpu);
          if not z80cpu.flagS then
          begin
            Inc(z80cpu.tStates); //ORH(4)
            Z80Push6T(z80cpu, z80cpu.regPC);
            z80cpu.regPC := lTemp;
          end;
        end;
      252: //CALL M,nn
        begin
          lTemp := LoadPCWord6T(z80cpu);
          if z80cpu.flagS then
          begin
            Inc(z80cpu.tStates); //ORH(4)
            Z80Push6T(z80cpu, z80cpu.regPC);
            z80cpu.regPC := lTemp;
          end;
        end;
      197: //PUSH BC
        begin
          Inc(z80cpu.tStates); //OCR(5)
          Z80Push6T(z80cpu, Z80GetBC(z80cpu));
        end;
      205: //CALL nn
        begin
          lTemp := LoadPCWord6T(z80cpu);
          Inc(z80cpu.tStates); //ORH(4)
          Z80Push6T(z80cpu, z80cpu.regPC);
          z80cpu.regPC := lTemp;
        end;
      213: //PUSH DE
        begin
          Inc(z80cpu.tStates); //OCR(5)
          Z80Push6T(z80cpu, z80cpu.regDE);
        end;
      221, 253: //prefix IX, prefix IY
        begin
          tmpB := Z80PeekBInternal(z80cpu, z80cpu.regPC);
            // "internal" to disable contention
            // ('cause double prefixes will be fetched and
            //  contended as usual)
          if (tmpB = 221) or (tmpB = 253) then continue;
            // double prefixing, just do nothing
          if tmpB in notIDOpSet then goto again; //continue;
            // the next instruction is not affected by ID prefix
            // so execute it as usual
            // note that it CAN'T be interrupted!
          prIX := (xxx = 221);
          // load next opcode (with the possible contention)
          Inc(z80cpu.tStates, 4); //OCR(4)
          xxx := Z80PeekB(z80cpu, z80cpu.regPC);
          Inc(z80cpu.regPC);
          Inc(z80cpu.regRTemp);

          if prIX then z80cpu.regID := z80cpu.regIX else z80cpu.regID := z80cpu.regIY;
          Z80ExecuteID(z80cpu, xxx);
          if prIX then z80cpu.regIX := z80cpu.regID else z80cpu.regIY := z80cpu.regID;
          //z80cpu.regID := 0; //tst!
        end;
      229: //PUSH HL
        begin
          Inc(z80cpu.tStates); //OCR(5)
          Z80Push6T(z80cpu, z80cpu.regHL);
        end;
      237: //prefix ED
        begin
          {$IFDEF USE_TRAPS}
          lTemp := Z80PeekBInternal(z80cpu, z80cpu.regPC);
          if Assigned(Z80OpTrapsED[lTemp]) then
          begin
            if Z80OpTrapsED[lTemp](z80cpu) then continue;
            // when this is called, #ED already fetched, PC & R incremented by 1,
            // t-states incremented by 4 (+possible contention)
            // return FALSE to continue as usual
            // return TRUE to fetch & run new instruction
            // set <breaked> to break immediately (res=TRUE) or after
            // interpretation of the current opcode (res=FALSE).
            // no state unrolling will be made on immediate break
          end;
          {$ENDIF}
          Z80ExecuteED(z80cpu);
        end;
      245: //PUSH AF
        begin
          Inc(z80cpu.tStates); //OCR(5)
          Z80Push6T(z80cpu, Z80GetAF(z80cpu));
        end;
      198: //ADD A,n
        Z80AddA(z80cpu, LoadPCByte3T(z80cpu));
      206: //ADC A,n
        Z80AdcA(z80cpu, LoadPCByte3T(z80cpu));
      214: //SUB n
        Z80SubA(z80cpu, LoadPCByte3T(z80cpu));
      222: //SBC n
        Z80SbcA(z80cpu, LoadPCByte3T(z80cpu));
      230: //AND n
        Z80AndA(z80cpu, LoadPCByte3T(z80cpu));
      238: //XOR n
        Z80XorA(z80cpu, LoadPCByte3T(z80cpu));
      246: //OR n
        Z80OrA(z80cpu, LoadPCByte3T(z80cpu));
      254: //CP n
        Z80CPA(z80cpu, LoadPCByte3T(z80cpu));
      199: //RST 0
        begin
          Inc(z80cpu.tStates); //OCR(5)
          Z80Push6T(z80cpu, z80cpu.regPC);
          z80cpu.regPC := 0;
        end;
      207: //RST 8
        begin
          Inc(z80cpu.tStates); //OCR(5)
          Z80Push6T(z80cpu, z80cpu.regPC);
          z80cpu.regPC := 8;
        end;
      215: //RST 16
        begin
          Inc(z80cpu.tStates); //OCR(5)
          Z80Push6T(z80cpu, z80cpu.regPC);
          z80cpu.regPC := 16;
        end;
      223: //RST 24
        begin
          Inc(z80cpu.tStates); //OCR(5)
          Z80Push6T(z80cpu, z80cpu.regPC);
          z80cpu.regPC := 24;
        end;
      231: //RST 32
        begin
          Inc(z80cpu.tStates); //OCR(5)
          Z80Push6T(z80cpu, z80cpu.regPC);
          z80cpu.regPC := 32;
        end;
      239: //RST 40
        begin
          Inc(z80cpu.tStates); //OCR(5)
          Z80Push6T(z80cpu, z80cpu.regPC);
          z80cpu.regPC := 40;
        end;
      247: //RST 48
        begin
          Inc(z80cpu.tStates); //OCR(5)
          Z80Push6T(z80cpu, z80cpu.regPC);
          z80cpu.regPC := 48;
        end;
      255: //RST 56
        begin
          Inc(z80cpu.tStates); //OCR(5)
          Z80Push6T(z80cpu, z80cpu.regPC);
          z80cpu.regPC := 56;
        end;
      end;
  end;
end;


procedure Z80Reset (z80cpu: PZ80CPU);
begin
  z80cpu.lastEItState := 0;
  z80cpu.regPC := 0;
  z80cpu.regSP := 0;
  z80cpu.regA := 0;
  Z80SetF(z80cpu, 0);
  Z80SetBC(z80cpu, 0);
  z80cpu.regDE := 0;
  z80cpu.regHL := 0;

  Z80EXX(z80cpu);
  Z80ExAFAF(z80cpu);

  z80cpu.regA := 0;
  Z80SetF(z80cpu, 0);
  Z80SetBC(z80cpu, 0);
  z80cpu.regDE := 0;
  z80cpu.regHL := 0;

  z80cpu.regIX := 0;
  z80cpu.regIY := 0;
  z80cpu.regR := 128;
  z80cpu.regRTemp := 0;

  z80cpu.regI := 0;
  z80cpu.iff1 := false;
  z80cpu.iff2 := false;
  z80cpu.im := 0;
  z80cpu.tStates := 0;

  z80cpu.halted := false;

  {$IFDEF USE_ONEMOREREAD}
  z80cpu.oneMoreRead := true;
  {$ENDIF}
  z80cpu.breaked := false;
  z80cpu.orgPC := 0;
  z80cpu.prevPC := 0;
end;

{$IFDEF USE_TRAPS}
procedure Z80ClearTraps ();
{$IFNDEF ONLY_ED_TRAPS}
var
  f: Integer;
begin
  for f := 0 to 255 do Z80OpTraps[f] := nil;
{$ELSE}
begin
{$ENDIF}
end;
{$ELSE}
procedure Z80ClearTraps ();
begin
end;
{$ENDIF}

procedure Z80ClearTrapsED ();
{$IFDEF USE_TRAPS}
var
  f: Integer;
begin
  for f := 0 to 255 do Z80OpTrapsED[f] := nil;
{$ELSE}
begin
{$ENDIF}
end;


procedure Z80InitParity ();
var
  f, c, mask: Integer;
  p: Boolean;
begin
  for f := 0 to 255 do
  begin
    p := true; mask := 1;
    for c := 0 to 7 do
    begin
      if (f and mask) <> 0 then p := not p;
      mask := mask shl 1;
    end;
    z80Parity[f] := p;
  end;
end;


begin
  Z80InitParity();
  Z80ClearTraps();
  Z80ClearTrapsED();
end.
